from collections.abc import Iterable
from typing import Generic, Self, TypeVar

import numpy as np

from pyautd3.driver.datagram.gain import Gain
from pyautd3.gain.holo.backend import Backend
from pyautd3.native_methods.autd3capi_gain_holo import EmissionConstraintWrap

from .amplitude import Amplitude

H = TypeVar("H", bound="Holo")


class Holo(Gain, Generic[H]):
    _foci: list[np.ndarray]
    _amps: list[Amplitude]
    _constraint: EmissionConstraintWrap

    def __init__(self: Self, constraint: EmissionConstraintWrap, iterable: Iterable[tuple[np.ndarray, Amplitude]]) -> None:
        foci = list(iterable)
        self._foci = [np.array(focus) for focus, _ in foci]
        self._amps = [amp for _, amp in foci]
        self._constraint = constraint

    def with_constraint(self: H, constraint: EmissionConstraintWrap) -> H:
        self._constraint = constraint
        return self


class HoloWithBackend(Holo[H], Generic[H]):
    _backend: Backend

    def __init__(
        self: Self,
        constraint: EmissionConstraintWrap,
        backend: Backend,
        iterable: Iterable[tuple[np.ndarray, Amplitude]],
    ) -> None:
        super().__init__(constraint, iterable)
        self._backend = backend



import asyncio
import ctypes
from collections.abc import Callable
from collections.abc import Iterable
from datetime import timedelta
from types import TracebackType
from typing import TYPE_CHECKING
from typing import Generic
from typing import Self
from typing import TypeVar
import numpy as np
from pyautd3.autd_error import InvalidDatagramTypeError
from pyautd3.autd_error import KeyAlreadyExistsError
from pyautd3.controller.timer import SpinSleeper
from pyautd3.controller.timer import TimerStrategy
from pyautd3.derive import builder
from pyautd3.driver.autd3_device import AUTD3
from pyautd3.driver.datagram import Datagram
from pyautd3.driver.firmware.fpga import FPGAState
from pyautd3.driver.firmware_version import FirmwareInfo
from pyautd3.driver.geometry import Device
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.link import Link
from pyautd3.driver.link import LinkBuilder
from pyautd3.native_methods.autd3capi import ControllerBuilderPtr
from pyautd3.native_methods.autd3capi import ControllerPtr
from pyautd3.native_methods.autd3capi import RuntimePtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import GeometryPtr
from pyautd3.native_methods.autd3capi_driver import HandlePtr
from pyautd3.native_methods.autd3capi_driver import TimerStrategyWrap
from pyautd3.native_methods.structs import FfiFuture
from pyautd3.native_methods.structs import Quaternion
from pyautd3.native_methods.structs import Vector3
from pyautd3.native_methods.utils import _validate_ptr
from pyautd3.native_methods.utils import _validate_status
from pyautd3.emulator import Emulator
from pyautd3.emulator import Recorder

K = TypeVar("K")
L = TypeVar("L", bound=Link)

class _Builder():
    devices: list[AUTD3]
    async def open_async(self: _Builder, link: LinkBuilder[L]) -> Controller[L]: ...
    def __init__(self: _Builder, iterable: Iterable[AUTD3]) -> None: ...
    def _ptr(self: _Builder, ) -> ControllerBuilderPtr: ...
    def open(self: _Builder, link: LinkBuilder[L]) -> Controller[L]: ...
    def into_emulator(self: _Builder, ) -> Emulator: ...
    def with_fallback_parallel_threshold(self: _Builder, fallback_parallel_threshold: int) -> _Builder: ...
    def with_fallback_timeout(self: _Builder, fallback_timeout: timedelta) -> _Builder: ...
    def with_send_interval(self: _Builder, send_interval: timedelta) -> _Builder: ...
    def with_receive_interval(self: _Builder, receive_interval: timedelta) -> _Builder: ...
    def with_timer_strategy(self: _Builder, timer_strategy: TimerStrategyWrap) -> _Builder: ...
    @property
    def fallback_parallel_threshold(self: _Builder) -> int: ...
    @property
    def fallback_timeout(self: _Builder) -> timedelta: ...
    @property
    def send_interval(self: _Builder) -> timedelta: ...
    @property
    def receive_interval(self: _Builder) -> timedelta: ...
    @property
    def timer_strategy(self: _Builder) -> TimerStrategyWrap: ...

class _GroupGuard(Generic[K]):
    _controller: Controller
    _keys: list[int]
    _datagrams: list[DatagramPtr]
    _keymap: dict[K, int]
    _k: int
    async def send_async(self: _GroupGuard[K], ) -> None: ...
    def __init__(self: _GroupGuard[K], group_map: Callable[[Device], K | None], controller: Controller) -> None: ...
    def set(self: _GroupGuard[K], key: K, d: Datagram | tuple[Datagram, Datagram]) -> _GroupGuard[K]: ...
    def send(self: _GroupGuard[K], ) -> None: ...

class Controller(Geometry, Generic[L]):
    _runtime: RuntimePtr
    _handle: HandlePtr
    _ptr: ControllerPtr
    _link: L
    _disposed: bool
    async def _open_impl_async(self: Controller[L], link_builder: LinkBuilder[L], timeout: timedelta | None = None) -> Controller[L]: ...
    async def firmware_version_async(self: Controller[L], ) -> list[FirmwareInfo]: ...
    async def close_async(self: Controller[L], ) -> None: ...
    async def fpga_state_async(self: Controller[L], ) -> list[FPGAState | None]: ...
    async def send_async(self: Controller[L], d: Datagram | tuple[Datagram, Datagram]) -> None: ...
    def __init__(self: Controller[L], geometry: GeometryPtr, runtime: RuntimePtr, handle: HandlePtr, ptr: ControllerPtr, link: L) -> None: ...
    def __del__(self: Controller[L], ) -> None: ...
    def _dispose(self: Controller[L], ) -> None: ...
    def __enter__(self: Controller[L], ) -> Controller[L]: ...
    def __exit__(self: Controller[L], _exc_type: type[BaseException] | None, _exc_value: BaseException | None, _traceback: TracebackType | None) -> None: ...
    def firmware_version(self: Controller[L], ) -> list[FirmwareInfo]: ...
    def close(self: Controller[L], ) -> None: ...
    def fpga_state(self: Controller[L], ) -> list[FPGAState | None]: ...
    def send(self: Controller[L], d: Datagram | tuple[Datagram, Datagram]) -> None: ...
    def group(self: Controller[L], group_map: Callable[[Device], K | None]) -> _GroupGuard: ...
    def tick(self: Controller[L], tick: timedelta) -> None: ...
    @staticmethod
    def builder(iterable: Iterable[AUTD3]) -> _Builder: ...
    @staticmethod
    def _open_impl(builder: ControllerBuilderPtr, link_builder: LinkBuilder[L], timeout: timedelta | None) -> Controller[L]: ...
    @property
    def link(self: Controller[L]) -> L: ...
    @property
    def geometry(self: Controller[L]) -> Geometry: ...



from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.derive import builder, datagram, gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Vector3


@datagram
@datagram_with_segment
@gain
@builder
class Focus(Gain):
    _prop_pos: np.ndarray
    _param_intensity: EmitIntensity
    _param_phase_offset: Phase

    def __init__(self: Self, pos: ArrayLike) -> None:
        super().__init__()
        self._prop_pos = np.array(pos)
        self._param_intensity = EmitIntensity.maximum()
        self._param_phase_offset = Phase(0)

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        return Base().gain_focus(
            Vector3(self._prop_pos),
            self._param_intensity.value,
            self._param_phase_offset.value,
        )



import contextlib

from .controller import Controller
from .driver.autd3_device import AUTD3
from .driver.datagram import (
    Clear,
    DebugSettings,
    DebugType,
    FixedCompletionTime,
    FixedUpdateRate,
    FociSTM,
    ForceFan,
    GainSTM,
    GainSTMMode,
    PhaseCorrection,
    PulseWidthEncoder,
    ReadsFPGAState,
    Silencer,
    SwapSegment,
)
from .driver.datagram.stm import (
    ControlPoint,
    ControlPoints1,
    ControlPoints2,
    ControlPoints3,
    ControlPoints4,
    ControlPoints5,
    ControlPoints6,
    ControlPoints7,
    ControlPoints8,
)
from .driver.defined import Hz, deg, kHz, rad
from .driver.firmware.fpga import Drive, EmitIntensity, LoopBehavior, Phase, SamplingConfig, TransitionMode
from .driver.geometry import Device, EulerAngles, Geometry, Transducer
from .ethercat import DcSysTime
from .gain import Bessel, Focus, Group, Null, Plane, Uniform
from .link.nop import Nop
from .modulation import Sine, Square, Static
from .native_methods.autd3capi import NativeMethods as Base
from .native_methods.autd3capi_driver import (
    GPIOIn,
    GPIOOut,
    Segment,
    SilencerTarget,
)
from .native_methods.autd3capi_emulator import NativeMethods as Emulator
from .native_methods.autd3capi_link_simulator import NativeMethods as Simulator
from .native_methods.autd3capi_link_twincat import NativeMethods as TwinCAT
from .native_methods.autd3capi_modulation_audio_file import NativeMethods as AudioFile


def tracing_init() -> None:
    Base().tracing_init()
    with contextlib.suppress(BaseException):
        Emulator().emulator_tracing_init()
    with contextlib.suppress(BaseException):
        Simulator().link_simulator_tracing_init()
    with contextlib.suppress(BaseException):
        TwinCAT().link_twin_cat_tracing_init()
    with contextlib.suppress(BaseException):
        AudioFile().modulation_audio_file_tracing_init()


__all__ = [
    "tracing_init",
    "Controller",
    "AUTD3",
    "PhaseCorrection",
    "Drive",
    "EmitIntensity",
    "Phase",
    "phase_rad",
    "SamplingConfig",
    "Clear",
    "Silencer",
    "FixedCompletionTime",
    "FixedUpdateRate",
    "DebugSettings",
    "DebugType",
    "ReadsFPGAState",
    "ForceFan",
    "ControlPoint",
    "ControlPoints1",
    "ControlPoints2",
    "ControlPoints3",
    "ControlPoints4",
    "ControlPoints5",
    "ControlPoints6",
    "ControlPoints7",
    "ControlPoints8",
    "FociSTM",
    "GainSTM",
    "GainSTMMode",
    "Device",
    "EulerAngles",
    "Geometry",
    "Transducer",
    "deg",
    "rad",
    "Bessel",
    "Focus",
    "Group",
    "Null",
    "Plane",
    "Custom",
    "Uniform",
    "Nop",
    "SamplingMode",
    "Sine",
    "Square",
    "Static",
    "LoopBehavior",
    "Segment",
    "SwapSegment",
    "GPIOIn",
    "GPIOOut",
    "Hz",
    "kHz",
    "DcSysTime",
    "PulseWidthEncoder",
    "TransitionMode",
    "SilencerTarget",
]

__version__ = "29.0.0rc4"



from typing import Self
import numpy as np
from numpy.typing import ArrayLike
from pyautd3.derive import builder
from pyautd3.derive import datagram
from pyautd3.derive import gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Vector3
from pyautd3.gain.cache import Cache
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class Plane(Gain):
    def __init__(self: Plane, direction: ArrayLike) -> None: ...
    def _gain_ptr(self: Plane, _: Geometry) -> GainPtr: ...
    def with_intensity(self: Plane, intensity: int | EmitIntensity) -> Plane: ...
    def with_phase_offset(self: Plane, phase_offset: int | Phase) -> Plane: ...
    def with_cache(self: Plane, ) -> Cache[Plane]: ...
    def with_timeout(self: Plane, timeout: timedelta | None) -> DatagramWithTimeout[Plane]: ...
    def with_parallel_threshold(self: Plane, threshold: int | None) -> DatagramWithParallelThreshold[Plane]: ...
    def with_segment(self: Plane, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Plane]: ...
    @property
    def dir(self: Plane) -> np.ndarray: ...
    @property
    def intensity(self: Plane) -> EmitIntensity: ...
    @property
    def phase_offset(self: Plane) -> Phase: ...



from abc import ABCMeta, abstractmethod
from typing import Self

from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi_driver import DatagramPtr

__all__ = []  # type: ignore[var-annotated]


class Datagram(metaclass=ABCMeta):
    @abstractmethod
    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        pass



# This file is autogenerated
import threading
import ctypes
import os
from pyautd3.native_methods.structs import Vector3, Quaternion, FfiFuture, LocalFfiFuture
from enum import IntEnum


class SyncMode(IntEnum):
    DC = 0
    FreeRun = 1

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class GainSTMMode(IntEnum):
    PhaseIntensityFull = 0
    PhaseFull = 1
    PhaseHalf = 2

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class GPIOOut(IntEnum):
    O0 = 0
    O1 = 1
    O2 = 2
    O3 = 3

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class GPIOIn(IntEnum):
    I0 = 0
    I1 = 1
    I2 = 2
    I3 = 3

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class Segment(IntEnum):
    S0 = 0
    S1 = 1

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class SilencerTarget(IntEnum):
    Intensity = 0
    PulseWidth = 1

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class Drive(ctypes.Structure):
    _fields_ = [("phase", ctypes.c_uint8), ("intensity", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Drive) and self._fields_ == other._fields_  # pragma: no cover


class LoopBehavior(ctypes.Structure):
    _fields_ = [("rep", ctypes.c_uint16)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, LoopBehavior) and self._fields_ == other._fields_  # pragma: no cover


class SamplingConfig(ctypes.Structure):
    _fields_ = [("division", ctypes.c_uint16)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SamplingConfig) and self._fields_ == other._fields_  # pragma: no cover





import ctypes
from collections.abc import Callable
from typing import Self
from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Device
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.geometry import Transducer
from pyautd3.ethercat.dc_sys_time import DcSysTime
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import DebugTypeWrap
from pyautd3.native_methods.autd3capi_driver import GeometryPtr
from pyautd3.native_methods.autd3capi_driver import GPIOOut
from pyautd3.native_methods.utils import ConstantADT
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold



class DebugType():
    NONE: DebugTypeWrap
    BaseSignal: DebugTypeWrap
    Thermo: DebugTypeWrap
    ForceFan: DebugTypeWrap
    Sync: DebugTypeWrap
    ModSegment: DebugTypeWrap
    StmSegment: DebugTypeWrap
    IsStmMode: DebugTypeWrap
    def __new__(cls, ) -> DebugType: ...
    @staticmethod
    def ModIdx(idx: int) -> DebugTypeWrap: ...
    @staticmethod
    def StmIdx(idx: int) -> DebugTypeWrap: ...
    @staticmethod
    def PwmOut(tr: Transducer) -> DebugTypeWrap: ...
    @staticmethod
    def Direct(value: bool) -> DebugTypeWrap: ...
    @staticmethod
    def SysTimeEq(value: DcSysTime) -> DebugTypeWrap: ...

class DebugSettings(Datagram):
    def __init__(self: DebugSettings, f: Callable[[Device, GPIOOut], DebugTypeWrap]) -> None: ...
    def _datagram_ptr(self: DebugSettings, geometry: Geometry) -> DatagramPtr: ...
    def with_timeout(self: DebugSettings, timeout: timedelta | None) -> DatagramWithTimeout[DebugSettings]: ...
    def with_parallel_threshold(self: DebugSettings, threshold: int | None) -> DatagramWithParallelThreshold[DebugSettings]: ...



from typing import Self

from pyautd3.driver.link import Link, LinkBuilder
from pyautd3.native_methods.autd3capi import ControllerPtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import HandlePtr, LinkBuilderPtr, LinkPtr
from pyautd3.native_methods.autd3capi_link_simulator import (
    NativeMethods as LinkSimulator,
)
from pyautd3.native_methods.utils import _to_null_terminated_utf8, _validate_ptr


class Simulator(Link):
    class _Builder(LinkBuilder["Simulator"]):
        addr: str

        def __init__(self: Self, addr: str) -> None:
            self.addr = addr

        def _link_builder_ptr(self: Self) -> LinkBuilderPtr:
            return _validate_ptr(LinkSimulator().link_simulator(_to_null_terminated_utf8(self.addr)))  # pragma: no cover

        def _resolve_link(self: Self, handle: HandlePtr, ptr: ControllerPtr) -> "Simulator":
            return Simulator(handle, Base().link_get(ptr))  # pragma: no cover

    def __init__(self: Self, handle: HandlePtr, ptr: LinkPtr) -> None:
        super().__init__(handle, ptr)  # pragma: no cover

    @staticmethod
    def builder(addr: str) -> _Builder:
        return Simulator._Builder(addr)



from datetime import timedelta
from pathlib import Path
from typing import Self

from pyautd3 import SamplingConfig
from pyautd3.derive import datagram, modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.defined.freq import Freq
from pyautd3.modulation.resample import Resampler
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.native_methods.autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile
from pyautd3.native_methods.utils import _to_null_terminated_utf8, _validate_ptr


@modulation
@datagram
@datagram_with_segment
class Wav(Modulation):
    _path: Path
    _resampler: tuple[SamplingConfig, Resampler] | None

    def __init__(self: Self, path: Path) -> None:
        super().__init__()
        self._path = path
        self._resampler = None

    @staticmethod
    def new_with_resample(
        path: Path,
        target: SamplingConfig | Freq[int] | Freq[float] | timedelta,
        resampler: Resampler,
    ) -> "Wav":
        instance = Wav(path)
        instance._resampler = (SamplingConfig(target), resampler)
        return instance

    def _modulation_ptr(self: Self) -> ModulationPtr:
        path = _to_null_terminated_utf8(str(self._path))
        match self._resampler:
            case (SamplingConfig(), Resampler()):
                (target, resampler) = self._resampler
                return _validate_ptr(
                    ModulationAudioFile().modulation_audio_file_wav_with_resample(
                        path,
                        self._loop_behavior,
                        target._inner,
                        resampler._dyn_resampler(),
                    ),
                )
            case _:
                return _validate_ptr(
                    ModulationAudioFile().modulation_audio_file_wav(
                        path,
                        self._loop_behavior,
                    ),
                )



from typing import Generic
from typing import Self
from typing import TypeVar
from pyautd3.derive import datagram
from pyautd3.derive import modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.modulation.cache import Cache
from pyautd3.modulation.fir import Fir
from collections.abc import Iterable
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment

M = TypeVar("M", bound=Modulation)

class RadiationPressure(Modulation, Generic[M]):
    _m: M
    def __init__(self: RadiationPressure[M], m: M) -> None: ...
    def _modulation_ptr(self: RadiationPressure[M], ) -> ModulationPtr: ...
    def with_cache(self: RadiationPressure[M], ) -> Cache[RadiationPressure[M]]: ...
    def with_fir(self: RadiationPressure[M], iterable: Iterable[float]) -> Fir[RadiationPressure[M]]: ...
    def with_timeout(self: RadiationPressure[M], timeout: timedelta | None) -> DatagramWithTimeout[RadiationPressure[M]]: ...
    def with_parallel_threshold(self: RadiationPressure[M], threshold: int | None) -> DatagramWithParallelThreshold[RadiationPressure[M]]: ...
    def with_segment(self: RadiationPressure[M], segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[RadiationPressure[M]]: ...



from typing import get_type_hints
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase


def builder(cls):
    annotations = get_type_hints(cls)

    for attr, ty in annotations.items():
        if attr.startswith("_param_"):
            if ty == EmitIntensity:
                prop_name = attr[7:]
                setter_name = f"with_{prop_name}"
                if not hasattr(cls, setter_name):

                    def setter(self, value, attr=attr):
                        from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity

                        value = EmitIntensity(value)
                        setattr(self, attr, value)
                        return self

                    setattr(cls, setter_name, setter)
            elif ty == Phase:
                prop_name = attr[7:]
                setter_name = f"with_{prop_name}"
                if not hasattr(cls, setter_name):

                    def setter(self, value, attr=attr):
                        from pyautd3.driver.firmware.fpga.phase import Phase

                        value = Phase(value)
                        setattr(self, attr, value)
                        return self

                    setattr(cls, setter_name, setter)
            elif attr.endswith("_u8"):
                prop_name = attr[7:-3]
                setter_name = f"with_{prop_name}"
                if not hasattr(cls, setter_name):

                    def setter(self, value, attr=attr):
                        from pyautd3.driver.utils import _validate_u8

                        value = _validate_u8(value)
                        setattr(self, attr, value)
                        return self

                    setattr(cls, setter_name, setter)
            else:
                prop_name = attr[7:]
                setter_name = f"with_{prop_name}"
                if not hasattr(cls, setter_name):

                    def setter(self, value, attr=attr):
                        setattr(self, attr, value)
                        return self

                    setattr(cls, setter_name, setter)

            if not hasattr(cls, prop_name):

                def getter(self, attr=attr):
                    return getattr(self, attr)

                prop = property(getter)
                setattr(cls, prop_name, prop)

        if attr.startswith("_prop_"):
            prop_name = attr[6:]
            if not hasattr(cls, prop_name):

                def getter(self, attr=attr):
                    return getattr(self, attr)

                prop = property(getter)
                setattr(cls, prop_name, prop)

    return cls



from typing import Self

from pyautd3.derive import datagram, modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.utils import _validate_u8
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr


@modulation
@datagram_with_segment
@datagram
class Static(Modulation):
    _intensity: int

    def __init__(self: Self, intensity: int | None = None) -> None:
        super().__init__()
        match intensity:
            case None:
                self._intensity = 0xFF
            case int():
                self._intensity = _validate_u8(intensity)
            case _:
                raise TypeError

    @staticmethod
    def with_intensity(intensity: int) -> "Static":
        return Static(intensity)

    @property
    def intensity(self: Self) -> int:
        return self._intensity

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return Base().modulation_static(self._intensity, self._loop_behavior)



from abc import ABCMeta, abstractmethod
from ctypes import Array, c_float
from typing import Self

from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.autd3capi_gain_holo import BackendPtr, EmissionConstraintWrap
from pyautd3.native_methods.structs import Vector3


class Backend(metaclass=ABCMeta):
    _ptr: BackendPtr

    def __init__(self: Self, ptr: BackendPtr) -> None:
        self._ptr = ptr

    def _backend_ptr(self: Self) -> BackendPtr:
        return self._ptr

    @abstractmethod
    def _gs(self: Self, foci: Array[Vector3], amps: Array[c_float], size: int, repeat: int, constraint: EmissionConstraintWrap) -> GainPtr:
        pass

    @abstractmethod
    def _gspat(self: Self, foci: Array[Vector3], amps: Array[c_float], size: int, repeat: int, constraint: EmissionConstraintWrap) -> GainPtr:
        pass

    @abstractmethod
    def _naive(self: Self, foci: Array[Vector3], amps: Array[c_float], size: int, constraint: EmissionConstraintWrap) -> GainPtr:
        pass

    @abstractmethod
    def _lm(
        self: Self,
        foci: Array[Vector3],
        amps: Array[c_float],
        size: int,
        eps1: float,
        eps2: float,
        tau: float,
        kmax: int,
        initial: Array[c_float],
        initial_size: int,
        constraint: EmissionConstraintWrap,
    ) -> GainPtr:
        pass



from typing import Self
import numpy as np
from numpy.typing import ArrayLike
from pyautd3.derive import builder
from pyautd3.derive import datagram
from pyautd3.derive import gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.defined.angle import Angle
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Vector3
from pyautd3.gain.cache import Cache
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class Bessel(Gain):
    def __init__(self: Bessel, pos: ArrayLike, direction: ArrayLike, theta: Angle) -> None: ...
    def _gain_ptr(self: Bessel, _: Geometry) -> GainPtr: ...
    def with_pos(self: Bessel, pos: np.ndarray) -> Bessel: ...
    def with_dir(self: Bessel, dir: np.ndarray) -> Bessel: ...
    def with_theta(self: Bessel, theta: Angle) -> Bessel: ...
    def with_intensity(self: Bessel, intensity: int | EmitIntensity) -> Bessel: ...
    def with_phase_offset(self: Bessel, phase_offset: int | Phase) -> Bessel: ...
    def with_cache(self: Bessel, ) -> Cache[Bessel]: ...
    def with_timeout(self: Bessel, timeout: timedelta | None) -> DatagramWithTimeout[Bessel]: ...
    def with_parallel_threshold(self: Bessel, threshold: int | None) -> DatagramWithParallelThreshold[Bessel]: ...
    def with_segment(self: Bessel, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Bessel]: ...
    @property
    def pos(self: Bessel) -> np.ndarray: ...
    @property
    def dir(self: Bessel) -> np.ndarray: ...
    @property
    def theta(self: Bessel) -> Angle: ...
    @property
    def intensity(self: Bessel) -> EmitIntensity: ...
    @property
    def phase_offset(self: Bessel) -> Phase: ...



from typing import Generic, Self, TypeVar

from pyautd3.derive import datagram, modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr

M = TypeVar("M", bound=Modulation)


@datagram
@datagram_with_segment
@modulation
class RadiationPressure(Modulation, Generic[M]):
    _m: M

    def __init__(self: Self, m: M) -> None:
        self._m = m
        self._loop_behavior = m._loop_behavior

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return Base().modulation_with_radiation_pressure(self._m._modulation_ptr(), self._loop_behavior)



import ctypes
from collections.abc import Callable
from typing import Self
from pyautd3.derive import datagram
from pyautd3.derive import gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga import Drive
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Device
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.geometry import Transducer
from pyautd3.native_methods.autd3_driver import Drive as _Drive
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ConstPtr
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.autd3capi_driver import GeometryPtr
from pyautd3.gain.cache import Cache
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class Custom(Gain):
    def __init__(self: Custom, f: Callable[[Device], Callable[[Transducer], Drive | EmitIntensity | Phase | tuple]]) -> None: ...
    def _gain_ptr(self: Custom, geometry: Geometry) -> GainPtr: ...
    def with_cache(self: Custom, ) -> Cache[Custom]: ...
    def with_timeout(self: Custom, timeout: timedelta | None) -> DatagramWithTimeout[Custom]: ...
    def with_parallel_threshold(self: Custom, threshold: int | None) -> DatagramWithParallelThreshold[Custom]: ...
    def with_segment(self: Custom, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Custom]: ...



# This file is autogenerated
import threading
import ctypes
import os
from pyautd3.native_methods.structs import Vector3, Quaternion, FfiFuture, LocalFfiFuture
from pyautd3.native_methods.autd3_driver import SamplingConfig, LoopBehavior, SyncMode, GainSTMMode, GPIOOut, GPIOIn, Segment, SilencerTarget, Drive
from pyautd3.native_methods.autd3capi_driver import GainPtr

from enum import IntEnum


class EmissionConstraintTag(IntEnum):
    Normalize = 1
    Uniform = 2
    Multiply = 3
    Clamp = 4

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class BackendPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class EmissionConstraintValue(ctypes.Union):
    _fields_ = [("null", ctypes.c_uint8), ("uniform", ctypes.c_uint8), ("multiply", ctypes.c_float), ("clamp", ctypes.c_uint8 * 2)]


class EmissionConstraintWrap(ctypes.Structure):
    _fields_ = [("tag", ctypes.c_uint8), ("value", EmissionConstraintValue)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, EmissionConstraintWrap) and self._fields_ == other._fields_  # pragma: no cover


class ResultBackend(ctypes.Structure):
    _fields_ = [("result", BackendPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultBackend) and self._fields_ == other._fields_  # pragma: no cover



class Singleton(type):
    _instances = {}  # type: ignore
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances: # pragma: no cover
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):

    def init_dll(self, bin_location: str, bin_prefix: str, bin_ext: str):
        self.dll = ctypes.CDLL(os.path.join(bin_location, f'{bin_prefix}autd3capi_gain_holo{bin_ext}'))

        self.dll.AUTDGainHoloConstraintNormalize.argtypes = [] 
        self.dll.AUTDGainHoloConstraintNormalize.restype = EmissionConstraintWrap

        self.dll.AUTDGainHoloConstraintUniform.argtypes = [ctypes.c_uint8] 
        self.dll.AUTDGainHoloConstraintUniform.restype = EmissionConstraintWrap

        self.dll.AUTDGainHoloConstraintMultiply.argtypes = [ctypes.c_float] 
        self.dll.AUTDGainHoloConstraintMultiply.restype = EmissionConstraintWrap

        self.dll.AUTDGainHoloConstraintClamp.argtypes = [ctypes.c_uint8, ctypes.c_uint8] 
        self.dll.AUTDGainHoloConstraintClamp.restype = EmissionConstraintWrap

        self.dll.AUTDGainHoloGreedySphere.argtypes = [ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, ctypes.c_uint8, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloGreedySphere.restype = GainPtr

        self.dll.AUTDGainHoloGreedyT4010A1.argtypes = [ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, ctypes.c_uint8, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloGreedyT4010A1.restype = GainPtr

        self.dll.AUTDGainGreedyIsDefault.argtypes = [EmissionConstraintWrap, ctypes.c_uint8]  # type: ignore 
        self.dll.AUTDGainGreedyIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainHoloGSSphere.argtypes = [BackendPtr, ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, ctypes.c_uint32, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloGSSphere.restype = GainPtr

        self.dll.AUTDGainHoloGST4010A1.argtypes = [BackendPtr, ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, ctypes.c_uint32, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloGST4010A1.restype = GainPtr

        self.dll.AUTDGainGSIsDefault.argtypes = [EmissionConstraintWrap, ctypes.c_uint32]  # type: ignore 
        self.dll.AUTDGainGSIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainHoloGSPATSphere.argtypes = [BackendPtr, ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, ctypes.c_uint32, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloGSPATSphere.restype = GainPtr

        self.dll.AUTDGainHoloGSPATT4010A1.argtypes = [BackendPtr, ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, ctypes.c_uint32, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloGSPATT4010A1.restype = GainPtr

        self.dll.AUTDGainGSPATIsDefault.argtypes = [EmissionConstraintWrap, ctypes.c_uint32]  # type: ignore 
        self.dll.AUTDGainGSPATIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainHoloSPLToPascal.argtypes = [ctypes.c_float] 
        self.dll.AUTDGainHoloSPLToPascal.restype = ctypes.c_float

        self.dll.AUTDGainHoloPascalToSPL.argtypes = [ctypes.c_float] 
        self.dll.AUTDGainHoloPascalToSPL.restype = ctypes.c_float

        self.dll.AUTDGainHoloLMSphere.argtypes = [BackendPtr, ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_uint32, ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloLMSphere.restype = GainPtr

        self.dll.AUTDGainHoloLMT4010A1.argtypes = [BackendPtr, ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_uint32, ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloLMT4010A1.restype = GainPtr

        self.dll.AUTDGainLMIsDefault.argtypes = [EmissionConstraintWrap, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_uint32, ctypes.POINTER(ctypes.c_float), ctypes.c_uint32]  # type: ignore 
        self.dll.AUTDGainLMIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainHoloNaiveSphere.argtypes = [BackendPtr, ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloNaiveSphere.restype = GainPtr

        self.dll.AUTDGainHoloNaiveT4010A1.argtypes = [BackendPtr, ctypes.POINTER(Vector3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainHoloNaiveT4010A1.restype = GainPtr

        self.dll.AUTDGainNaiveIsDefault.argtypes = [EmissionConstraintWrap]  # type: ignore 
        self.dll.AUTDGainNaiveIsDefault.restype = ctypes.c_bool

        self.dll.AUTDNalgebraBackendSphere.argtypes = [] 
        self.dll.AUTDNalgebraBackendSphere.restype = BackendPtr

        self.dll.AUTDNalgebraBackendT4010A1.argtypes = [] 
        self.dll.AUTDNalgebraBackendT4010A1.restype = BackendPtr

        self.dll.AUTDDeleteNalgebraBackendSphere.argtypes = [BackendPtr]  # type: ignore 
        self.dll.AUTDDeleteNalgebraBackendSphere.restype = None

        self.dll.AUTDDeleteNalgebraBackendT4010A1.argtypes = [BackendPtr]  # type: ignore 
        self.dll.AUTDDeleteNalgebraBackendT4010A1.restype = None

    def gain_holo_constraint_normalize(self) -> EmissionConstraintWrap:
        return self.dll.AUTDGainHoloConstraintNormalize()

    def gain_holo_constraint_uniform(self, intensity: int) -> EmissionConstraintWrap:
        return self.dll.AUTDGainHoloConstraintUniform(intensity)

    def gain_holo_constraint_multiply(self, v: float) -> EmissionConstraintWrap:
        return self.dll.AUTDGainHoloConstraintMultiply(v)

    def gain_holo_constraint_clamp(self, min_v: int, max_v: int) -> EmissionConstraintWrap:
        return self.dll.AUTDGainHoloConstraintClamp(min_v, max_v)

    def gain_holo_greedy_sphere(self, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, div: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloGreedySphere(points, amps, size, div, constraint)

    def gain_holo_greedy_t_4010_a_1(self, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, div: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloGreedyT4010A1(points, amps, size, div, constraint)  # pragma: no cover

    def gain_greedy_is_default(self, constraint: EmissionConstraintWrap, phase_div: int) -> ctypes.c_bool:
        return self.dll.AUTDGainGreedyIsDefault(constraint, phase_div)

    def gain_holo_gs_sphere(self, backend: BackendPtr, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, repeat_nonzero: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloGSSphere(backend, points, amps, size, repeat_nonzero, constraint)

    def gain_holo_gst_4010_a_1(self, backend: BackendPtr, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, repeat_nonzero: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloGST4010A1(backend, points, amps, size, repeat_nonzero, constraint)  # pragma: no cover

    def gain_gs_is_default(self, constraint: EmissionConstraintWrap, repeat: int) -> ctypes.c_bool:
        return self.dll.AUTDGainGSIsDefault(constraint, repeat)

    def gain_holo_gspat_sphere(self, backend: BackendPtr, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, repeat_nonzero: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloGSPATSphere(backend, points, amps, size, repeat_nonzero, constraint)

    def gain_holo_gspatt_4010_a_1(self, backend: BackendPtr, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, repeat_nonzero: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloGSPATT4010A1(backend, points, amps, size, repeat_nonzero, constraint)  # pragma: no cover

    def gain_gspat_is_default(self, constraint: EmissionConstraintWrap, repeat: int) -> ctypes.c_bool:
        return self.dll.AUTDGainGSPATIsDefault(constraint, repeat)

    def gain_holo_spl_to_pascal(self, value: float) -> ctypes.c_float:
        return self.dll.AUTDGainHoloSPLToPascal(value)

    def gain_holo_pascal_to_spl(self, value: float) -> ctypes.c_float:
        return self.dll.AUTDGainHoloPascalToSPL(value)

    def gain_holo_lm_sphere(self, backend: BackendPtr, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, eps_1: float, eps_2: float, tau: float, k_max_nonzero: int, initial_ptr: ctypes.Array[ctypes.c_float] | None, initial_len: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloLMSphere(backend, points, amps, size, eps_1, eps_2, tau, k_max_nonzero, initial_ptr, initial_len, constraint)

    def gain_holo_lmt_4010_a_1(self, backend: BackendPtr, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, eps_1: float, eps_2: float, tau: float, k_max_nonzero: int, initial_ptr: ctypes.Array[ctypes.c_float] | None, initial_len: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloLMT4010A1(backend, points, amps, size, eps_1, eps_2, tau, k_max_nonzero, initial_ptr, initial_len, constraint)  # pragma: no cover

    def gain_lm_is_default(self, constraint: EmissionConstraintWrap, eps_1: float, eps_2: float, tau: float, k_max: int, initial_ptr: ctypes.Array[ctypes.c_float] | None, initial_len: int) -> ctypes.c_bool:
        return self.dll.AUTDGainLMIsDefault(constraint, eps_1, eps_2, tau, k_max, initial_ptr, initial_len)

    def gain_holo_naive_sphere(self, backend: BackendPtr, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloNaiveSphere(backend, points, amps, size, constraint)

    def gain_holo_naive_t_4010_a_1(self, backend: BackendPtr, points: ctypes.Array | None, amps: ctypes.Array[ctypes.c_float] | None, size: int, constraint: EmissionConstraintWrap) -> GainPtr:
        return self.dll.AUTDGainHoloNaiveT4010A1(backend, points, amps, size, constraint)  # pragma: no cover

    def gain_naive_is_default(self, constraint: EmissionConstraintWrap) -> ctypes.c_bool:
        return self.dll.AUTDGainNaiveIsDefault(constraint)

    def nalgebra_backend_sphere(self) -> BackendPtr:
        return self.dll.AUTDNalgebraBackendSphere()

    def nalgebra_backend_t_4010_a_1(self) -> BackendPtr:
        return self.dll.AUTDNalgebraBackendT4010A1()  # pragma: no cover

    def delete_nalgebra_backend_sphere(self, backend: BackendPtr) -> None:
        return self.dll.AUTDDeleteNalgebraBackendSphere(backend)

    def delete_nalgebra_backend_t_4010_a_1(self, backend: BackendPtr) -> None:
        return self.dll.AUTDDeleteNalgebraBackendT4010A1(backend)  # pragma: no cover



from datetime import timedelta
from typing import Self

from pyautd3.native_methods.autd3capi import NativeMethods as Base


class DcSysTime:
    _dc_sys_time: int

    def __new__(cls: type["DcSysTime"]) -> "DcSysTime":
        raise NotImplementedError

    @classmethod
    def __private_new__(cls: type["DcSysTime"], dc_sys_time: int) -> "DcSysTime":
        ins = super().__new__(cls)
        ins._dc_sys_time = dc_sys_time
        return ins

    @property
    def sys_time(self: Self) -> int:
        return self._dc_sys_time

    @staticmethod
    def now() -> "DcSysTime":
        return DcSysTime.__private_new__(int(Base().dc_sys_time_now()))

    def __add__(self: Self, other: timedelta) -> "DcSysTime":
        return DcSysTime.__private_new__(self._dc_sys_time + int(other.total_seconds() * 1000 * 1000 * 1000))

    def __sub__(self: Self, other: "timedelta") -> "DcSysTime":
        return DcSysTime.__private_new__(self._dc_sys_time - int(other.total_seconds() * 1000 * 1000 * 1000))



import ctypes
from abc import ABCMeta, abstractmethod
from typing import Self

import numpy as np

from pyautd3.driver.defined.angle import Angle
from pyautd3.driver.defined.freq import Freq
from pyautd3.native_methods.autd3_driver import SamplingConfig
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import LoopBehavior, ModulationPtr
from pyautd3.native_methods.utils import _validate_ptr


class ISamplingMode(metaclass=ABCMeta):
    @abstractmethod
    def sine_ptr(
        self: Self,
        config: SamplingConfig,
        intensity: int,
        offset: int,
        phase: Angle,
        clamp: bool,  # noqa: FBT001
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        pass

    @abstractmethod
    def sine_freq(self: Self) -> int | float:
        pass

    @abstractmethod
    def fourier_ptr(
        self: Self,
        components: list,
        size: int,
        clamp: bool,  # noqa: FBT001
        scale_factor: float,
        offset: int,
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        pass

    @abstractmethod
    def square_ptr(
        self: Self,
        config: SamplingConfig,
        low: int,
        high: int,
        duty: float,
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        pass

    @abstractmethod
    def square_freq(self: Self) -> int | float:
        pass


class SamplingModeExact(ISamplingMode):
    _freq: Freq[int]

    def __init__(self: Self, freq: Freq[int]) -> None:
        self._freq = freq

    def sine_ptr(
        self: Self,
        config: SamplingConfig,
        intensity: int,
        offset: int,
        phase: Angle,
        clamp: bool,  # noqa: FBT001
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        return _validate_ptr(
            Base().modulation_sine_exact(
                self._freq.hz,
                config,
                intensity,
                offset,
                phase.radian,
                clamp,
                loop_behavior,
            ),
        )

    def sine_freq(self: Self) -> int | float:
        return int(Base().modulation_sine_exact_freq(self._freq.hz))

    def fourier_ptr(
        self: Self,
        components: list,
        size: int,
        clamp: bool,  # noqa: FBT001
        scale_factor: float,
        offset: int,
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        sine_freq = np.fromiter((m._mode._freq.hz for m in components), dtype=np.uint32)
        sine_config = np.fromiter((np.void(m._config._inner) for m in components), dtype=SamplingConfig)
        sine_intensity = np.fromiter((m._param_intensity_u8 for m in components), dtype=np.uint8)
        sine_offset = np.fromiter((m._param_offset_u8 for m in components), dtype=np.uint8)
        sine_phase = np.fromiter((m._param_phase.radian for m in components), dtype=np.float32)
        sine_clamp = np.fromiter((m._param_clamp for m in components), dtype=np.bool_)
        return _validate_ptr(
            Base().modulation_fourier_exact(
                sine_freq.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32)),  # type: ignore[arg-type]
                sine_config.ctypes.data_as(ctypes.POINTER(SamplingConfig)),  # type: ignore[arg-type]
                sine_intensity.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                sine_offset.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                sine_phase.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                sine_clamp.ctypes.data_as(ctypes.POINTER(ctypes.c_bool)),  # type: ignore[arg-type]
                size,
                clamp,
                scale_factor,
                offset,
                loop_behavior,
            ),
        )

    def square_ptr(
        self: Self,
        config: SamplingConfig,
        low: int,
        high: int,
        duty: float,
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        return _validate_ptr(
            Base().modulation_square_exact(
                self._freq.hz,
                config,
                low,
                high,
                duty,
                loop_behavior,
            ),
        )

    def square_freq(self: Self) -> int | float:
        return int(Base().modulation_square_exact_freq(self._freq.hz))


class SamplingModeExactFloat(ISamplingMode):
    _freq: Freq[float]

    def __init__(self: Self, freq: Freq[float]) -> None:
        self._freq = freq

    def sine_ptr(
        self: Self,
        config: SamplingConfig,
        intensity: int,
        offset: int,
        phase: Angle,
        clamp: bool,  # noqa: FBT001
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        return _validate_ptr(
            Base().modulation_sine_exact_float(
                self._freq.hz,
                config,
                intensity,
                offset,
                phase.radian,
                clamp,
                loop_behavior,
            ),
        )

    def sine_freq(self: Self) -> float:
        return float(Base().modulation_sine_exact_float_freq(self._freq.hz))

    def fourier_ptr(
        self: Self,
        components: list,
        size: int,
        clamp: bool,  # noqa: FBT001
        scale_factor: float,
        offset: int,
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        sine_freq = np.fromiter((m._mode._freq.hz for m in components), dtype=np.float32)
        sine_config = np.fromiter((np.void(m._config._inner) for m in components), dtype=SamplingConfig)
        sine_intensity = np.fromiter((m._param_intensity_u8 for m in components), dtype=np.uint8)
        sine_offset = np.fromiter((m._param_offset_u8 for m in components), dtype=np.uint8)
        sine_phase = np.fromiter((m._param_phase.radian for m in components), dtype=np.float32)
        sine_clamp = np.fromiter((m._param_clamp for m in components), dtype=np.bool_)
        return _validate_ptr(
            Base().modulation_fourier_exact_float(
                sine_freq.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                sine_config.ctypes.data_as(ctypes.POINTER(SamplingConfig)),  # type: ignore[arg-type]
                sine_intensity.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                sine_offset.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                sine_phase.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                sine_clamp.ctypes.data_as(ctypes.POINTER(ctypes.c_bool)),  # type: ignore[arg-type]
                size,
                clamp,
                scale_factor,
                offset,
                loop_behavior,
            ),
        )

    def square_ptr(
        self: Self,
        config: SamplingConfig,
        low: int,
        high: int,
        duty: float,
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        return _validate_ptr(
            Base().modulation_square_exact_float(
                self._freq.hz,
                config,
                low,
                high,
                duty,
                loop_behavior,
            ),
        )

    def square_freq(self: Self) -> float:
        return int(Base().modulation_square_exact_float_freq(self._freq.hz))


class SamplingModeNearest(ISamplingMode):
    _freq: Freq[float]

    def __init__(self: Self, freq: Freq[float]) -> None:
        self._freq = freq

    def sine_ptr(
        self: Self,
        config: SamplingConfig,
        intensity: int,
        offset: int,
        phase: Angle,
        clamp: bool,  # noqa: FBT001
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        return _validate_ptr(
            Base().modulation_sine_nearest(
                self._freq.hz,
                config,
                intensity,
                offset,
                phase.radian,
                clamp,
                loop_behavior,
            ),
        )

    def sine_freq(self: Self) -> int | float:
        return float(Base().modulation_sine_nearest_freq(self._freq.hz))

    def fourier_ptr(
        self: Self,
        components: list,
        size: int,
        clamp: bool,  # noqa: FBT001
        scale_factor: float,
        offset: int,
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        sine_freq = np.fromiter((m._mode._freq.hz for m in components), dtype=np.float32)
        sine_config = np.fromiter((np.void(m._config._inner) for m in components), dtype=SamplingConfig)
        sine_intensity = np.fromiter((m._param_intensity_u8 for m in components), dtype=np.uint8)
        sine_offset = np.fromiter((m._param_offset_u8 for m in components), dtype=np.uint8)
        sine_phase = np.fromiter((m._param_phase.radian for m in components), dtype=np.float32)
        sine_clamp = np.fromiter((m._param_clamp for m in components), dtype=np.bool_)
        return _validate_ptr(
            Base().modulation_fourier_nearest(
                sine_freq.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                sine_config.ctypes.data_as(ctypes.POINTER(SamplingConfig)),  # type: ignore[arg-type]
                sine_intensity.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                sine_offset.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                sine_phase.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                sine_clamp.ctypes.data_as(ctypes.POINTER(ctypes.c_bool)),  # type: ignore[arg-type]
                size,
                clamp,
                scale_factor,
                offset,
                loop_behavior,
            ),
        )

    def square_ptr(
        self: Self,
        config: SamplingConfig,
        low: int,
        high: int,
        duty: float,
        loop_behavior: LoopBehavior,
    ) -> ModulationPtr:
        return _validate_ptr(
            Base().modulation_square_nearest(
                self._freq.hz,
                config,
                low,
                high,
                duty,
                loop_behavior,
            ),
        )

    def square_freq(self: Self) -> int | float:
        return int(Base().modulation_square_nearest_freq(self._freq.hz))



import os
import os.path
import platform
import sys

from .autd3capi import NativeMethods as Base
from .autd3capi_gain_holo import NativeMethods as GainHolo
from .autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile
from .autd3capi_link_simulator import NativeMethods as LinkSimulator
from .autd3capi_link_twincat import NativeMethods as LinkTwincAT
from .autd3capi_emulator import NativeMethods as Emulator

_PLATFORM = platform.system()
_PREFIX = ""
_BIN_EXT = ""
if _PLATFORM == "Windows":
    _BIN_EXT = ".dll"
elif _PLATFORM == "Darwin":
    _PREFIX = "lib"
    _BIN_EXT = ".dylib"
elif _PLATFORM == "Linux":
    _PREFIX = "lib"
    _BIN_EXT = ".so"
else:
    raise ImportError("Not supported OS")

_LIB_PATH = os.path.join(os.path.dirname(__file__), "..", "bin")

Base().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
GainHolo().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
ModulationAudioFile().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
LinkSimulator().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
if sys.platform == "win32":
    try:
        os.add_dll_directory("C:\\TwinCAT\\Common64")
    except FileNotFoundError:
        pass
try:
    LinkTwincAT().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
except Exception:  # pragma: no cover
    pass  # pragma: no cover
Emulator().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)



from datetime import timedelta
from typing import Self

from pyautd3.driver.defined import Freq
from pyautd3.driver.defined.freq import Hz
from pyautd3.driver.utils import _validate_nonzero_u16
from pyautd3.native_methods.autd3_driver import SamplingConfig as _SamplingConfig
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.utils import _validate_sampling_config


class SamplingConfig:
    _inner: _SamplingConfig

    def __init__(self: Self, value: "_SamplingConfig | SamplingConfig | int | Freq[int] | Freq[float] | timedelta") -> None:
        match value:
            case int():
                self._inner = _validate_sampling_config(Base().sampling_config_from_division(_validate_nonzero_u16(value)))
            case _SamplingConfig():
                self._inner = value
            case SamplingConfig():
                self._inner = value._inner
            case Freq():
                match value.hz:
                    case int():
                        self._inner = _validate_sampling_config(Base().sampling_config_from_freq(value.hz))
                    case float():
                        self._inner = _validate_sampling_config(Base().sampling_config_from_freq_f(value.hz))
                    case _:
                        raise TypeError
            case timedelta():
                self._inner = _validate_sampling_config(Base().sampling_config_from_period(int(value.total_seconds() * 1000 * 1000 * 1000)))
            case _:
                raise TypeError

    @staticmethod
    def nearest(value: Freq[float] | timedelta) -> "SamplingConfig":
        match value:
            case Freq():
                match value.hz:
                    case float():
                        return SamplingConfig(Base().sampling_config_from_freq_nearest(value.hz))
                    case _:
                        raise TypeError
            case timedelta():
                return SamplingConfig(Base().sampling_config_from_period_nearest(int(value.total_seconds() * 1000 * 1000 * 1000)))
            case _:
                raise TypeError

    @property
    def division(self: Self) -> int:
        return int(Base().sampling_config_division(self._inner))

    @property
    def freq(self: Self) -> Freq[float]:
        return float(Base().sampling_config_freq(self._inner)) * Hz

    @property
    def period(self: Self) -> timedelta:
        return timedelta(microseconds=int(Base().sampling_config_period(self._inner)) / 1000)

    def __eq__(self: Self, value: object) -> bool:
        return isinstance(value, SamplingConfig) and self._inner.division == self._inner.division



import ctypes
from typing import Self


class AUTDError(Exception):
    msg: str

    def __init__(self: Self, err: ctypes.Array[ctypes.c_char] | str) -> None:
        self.msg = err if isinstance(err, str) else err.value.decode("utf-8")

    def __str__(self: Self) -> str:
        return self.msg

    def __repr__(self: Self) -> str:
        return self.msg


class UnknownGroupKeyError(AUTDError):
    def __init__(self: Self) -> None:
        super().__init__("Unknown group key")


class KeyAlreadyExistsError(AUTDError):
    def __init__(self: Self) -> None:
        super().__init__("Key already exists")


class InvalidDatagramTypeError(AUTDError):
    def __init__(self: Self) -> None:
        super().__init__("Invalid datagram type")


class CantBeZeroError(AUTDError):
    def __init__(self: Self, v: int) -> None:
        super().__init__(f"Value must be greater than 0: {v}")



import ctypes
from collections.abc import Callable
from typing import Self
from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Device
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import GeometryPtr
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold



class ReadsFPGAState(Datagram):
    def __init__(self: ReadsFPGAState, f: Callable[[Device], bool]) -> None: ...
    def _datagram_ptr(self: ReadsFPGAState, geometry: Geometry) -> DatagramPtr: ...
    def with_timeout(self: ReadsFPGAState, timeout: timedelta | None) -> DatagramWithTimeout[ReadsFPGAState]: ...
    def with_parallel_threshold(self: ReadsFPGAState, threshold: int | None) -> DatagramWithParallelThreshold[ReadsFPGAState]: ...



from typing import Generic
from typing import Self
from typing import TypeVar
from pyautd3.derive import datagram
from pyautd3.derive import gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import GainCachePtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment

G = TypeVar("G", bound=Gain)

class Cache(Gain, Generic[G]):
    _g: G
    _ptr: GainCachePtr | None
    def __init__(self: Cache[G], g: G) -> None: ...
    def _gain_ptr(self: Cache[G], geometry: Geometry) -> GainPtr: ...
    def __del__(self: Cache[G], ) -> None: ...
    def with_timeout(self: Cache[G], timeout: timedelta | None) -> DatagramWithTimeout[Cache[G]]: ...
    def with_parallel_threshold(self: Cache[G], threshold: int | None) -> DatagramWithParallelThreshold[Cache[G]]: ...
    def with_segment(self: Cache[G], segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Cache[G]]: ...



# This file is autogenerated
import threading
import ctypes
import os
from pyautd3.native_methods.structs import Vector3, Quaternion, FfiFuture, LocalFfiFuture
from pyautd3.native_methods.autd3_driver import SamplingConfig, LoopBehavior, SyncMode, GainSTMMode, GPIOOut, GPIOIn, Segment, SilencerTarget, Drive
from pyautd3.native_methods.autd3capi_driver import ResultStatus, ResultSyncLinkBuilder



class Singleton(type):
    _instances = {}  # type: ignore
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances: # pragma: no cover
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):

    def init_dll(self, bin_location: str, bin_prefix: str, bin_ext: str):
        self.dll = ctypes.CDLL(os.path.join(bin_location, f'{bin_prefix}autd3capi_link_simulator{bin_ext}'))

        self.dll.AUTDLinkSimulatorTracingInit.argtypes = [] 
        self.dll.AUTDLinkSimulatorTracingInit.restype = None

        self.dll.AUTDLinkSimulatorTracingInitWithFile.argtypes = [ctypes.c_char_p] 
        self.dll.AUTDLinkSimulatorTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDLinkSimulator.argtypes = [ctypes.c_char_p] 
        self.dll.AUTDLinkSimulator.restype = ResultSyncLinkBuilder

    def link_simulator_tracing_init(self) -> None:
        return self.dll.AUTDLinkSimulatorTracingInit()

    def link_simulator_tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDLinkSimulatorTracingInitWithFile(path)

    def link_simulator(self, addr: bytes) -> ResultSyncLinkBuilder:
        return self.dll.AUTDLinkSimulator(addr)



import asyncio
import ctypes
from collections.abc import Callable, Iterable
from datetime import timedelta
from types import TracebackType
from typing import TYPE_CHECKING, Generic, Self, TypeVar

import numpy as np

from pyautd3.autd_error import InvalidDatagramTypeError, KeyAlreadyExistsError
from pyautd3.controller.timer import SpinSleeper, TimerStrategy
from pyautd3.derive import builder
from pyautd3.driver.autd3_device import AUTD3
from pyautd3.driver.datagram import Datagram
from pyautd3.driver.firmware.fpga import FPGAState
from pyautd3.driver.firmware_version import FirmwareInfo
from pyautd3.driver.geometry import Device, Geometry
from pyautd3.driver.link import Link, LinkBuilder
from pyautd3.native_methods.autd3capi import ControllerBuilderPtr, ControllerPtr, RuntimePtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GeometryPtr, HandlePtr, TimerStrategyWrap
from pyautd3.native_methods.structs import FfiFuture, Quaternion, Vector3
from pyautd3.native_methods.utils import _validate_ptr, _validate_status

if TYPE_CHECKING:
    from pyautd3.emulator import Emulator, Recorder

K = TypeVar("K")
L = TypeVar("L", bound=Link)


@builder
class _Builder:
    devices: list[AUTD3]
    _param_fallback_parallel_threshold: int
    _param_fallback_timeout: timedelta
    _param_send_interval: timedelta
    _param_receive_interval: timedelta
    _param_timer_strategy: TimerStrategyWrap

    def __init__(self: Self, iterable: Iterable[AUTD3]) -> None:
        self.devices = list(iterable)
        self._param_fallback_parallel_threshold = 4
        self._param_fallback_timeout = timedelta(milliseconds=20)
        self._param_send_interval = timedelta(milliseconds=1)
        self._param_receive_interval = timedelta(milliseconds=1)
        self._param_timer_strategy = TimerStrategy.Spin(SpinSleeper())

    def _ptr(self: Self) -> ControllerBuilderPtr:
        pos = np.fromiter((np.void(Vector3(d.position)) for d in self.devices), dtype=Vector3)  # type: ignore[type-var,call-overload]
        rot = np.fromiter((np.void(Quaternion(d.rotation)) for d in self.devices), dtype=Quaternion)  # type: ignore[type-var,call-overload]
        return Base().controller_builder(
            pos.ctypes.data_as(ctypes.POINTER(Vector3)),  # type: ignore[arg-type]
            rot.ctypes.data_as(ctypes.POINTER(Quaternion)),  # type: ignore[arg-type]
            len(pos),
            self._param_fallback_parallel_threshold,
            int(self._param_fallback_timeout.total_seconds() * 1000 * 1000 * 1000),
            int(self._param_send_interval.total_seconds() * 1000 * 1000 * 1000),
            int(self._param_receive_interval.total_seconds() * 1000 * 1000 * 1000),
            self._param_timer_strategy,
        )

    async def open_async(
        self: Self,
        link: LinkBuilder[L],
        *,
        timeout: timedelta | None = None,  # noqa: ASYNC109
    ) -> "Controller[L]":
        return await Controller._open_impl_async(self._ptr(), link, timeout)

    def open(self: Self, link: LinkBuilder[L], *, timeout: timedelta | None = None) -> "Controller[L]":
        return Controller._open_impl(self._ptr(), link, timeout)

    def into_emulator(self: Self) -> "Emulator":  # type: ignore[empty-body]
        pass  # pragma: no cover


class _GroupGuard(Generic[K]):
    _controller: "Controller"
    _keys: list[int]
    _datagrams: list[DatagramPtr]
    _keymap: dict[K, int]
    _k: int

    def __init__(self: Self, group_map: Callable[[Device], K | None], controller: "Controller") -> None:
        def f_native(_context: ctypes.c_void_p, geometry_ptr: GeometryPtr, dev_idx: int) -> int:
            dev = Device(dev_idx, geometry_ptr)
            key = group_map(dev)
            return self._keymap[key] if key is not None else -1

        self._f_native = ctypes.CFUNCTYPE(ctypes.c_int32, ctypes.c_void_p, GeometryPtr, ctypes.c_uint16)(f_native)
        self._controller = controller
        self._keys = []
        self._datagrams = []
        self._keymap = {}
        self._k = 0

    def set(
        self: Self,
        key: K,
        d: Datagram | tuple[Datagram, Datagram],
    ) -> "_GroupGuard[K]":
        if key in self._keymap:
            raise KeyAlreadyExistsError
        match d:
            case Datagram():
                ptr = d._datagram_ptr(self._controller.geometry)
                self._datagrams.append(ptr)
            case (Datagram(), Datagram()):
                (d1, d2) = d
                ptr = Base().datagram_tuple(
                    d1._datagram_ptr(self._controller.geometry),
                    d2._datagram_ptr(self._controller.geometry),
                )
                self._datagrams.append(ptr)
            case _:
                raise InvalidDatagramTypeError
        self._keys.append(self._k)
        self._keymap[key] = self._k
        self._k += 1

        return self

    async def send_async(self: Self) -> None:
        keys = np.array(self._keys, dtype=np.int32)
        datagrams: np.ndarray = np.ndarray(len(self._datagrams), dtype=DatagramPtr)
        for i, d in enumerate(self._datagrams):
            datagrams[i]["_0"] = d._0
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()
        ffi_future = Base().controller_group(
            self._controller._ptr,
            self._f_native,  # type: ignore[arg-type]
            None,
            self._controller._geometry_ptr,
            keys.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),  # type: ignore[arg-type]
            datagrams.ctypes.data_as(ctypes.POINTER(DatagramPtr)),  # type: ignore[arg-type]
            len(self._keys),
        )
        loop.call_soon(
            lambda *_: future.set_result(
                Base().wait_local_result_status(self._controller._handle, ffi_future),
            ),
        )
        _validate_status(await future)

    def send(self: Self) -> None:
        keys = np.array(self._keys, dtype=np.int32)
        datagrams: np.ndarray = np.ndarray(len(self._datagrams), dtype=DatagramPtr)
        for i, d in enumerate(self._datagrams):
            datagrams[i]["_0"] = d._0
        _validate_status(
            Base().wait_local_result_status(
                self._controller._handle,
                Base().controller_group(
                    self._controller._ptr,
                    self._f_native,  # type: ignore[arg-type]
                    None,
                    self._controller._geometry_ptr,
                    keys.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),  # type: ignore[arg-type]
                    datagrams.ctypes.data_as(ctypes.POINTER(DatagramPtr)),  # type: ignore[arg-type]
                    len(self._keys),
                ),
            ),
        )


class Controller(Geometry, Generic[L]):
    _runtime: RuntimePtr
    _handle: HandlePtr
    _ptr: ControllerPtr
    _link: L
    _disposed: bool

    def __init__(self: Self, geometry: GeometryPtr, runtime: RuntimePtr, handle: HandlePtr, ptr: ControllerPtr, link: L) -> None:
        super().__init__(geometry)
        self._runtime = runtime
        self._handle = handle
        self._ptr = ptr
        self._link = link
        self._disposed = False

    @property
    def link(self: Self) -> L:
        return self._link

    @staticmethod
    def builder(iterable: Iterable[AUTD3]) -> _Builder:
        return _Builder(iterable)

    def __del__(self: Self) -> None:
        self._dispose()

    def _dispose(self: Self) -> None:
        self.close()

    def __enter__(self: Self) -> "Controller[L]":
        return self

    def __exit__(
        self: Self,
        _exc_type: type[BaseException] | None,
        _exc_value: BaseException | None,
        _traceback: TracebackType | None,
    ) -> None:
        self._dispose()

    @property
    def geometry(self: Self) -> Geometry:
        return self  # type: ignore[return-value]

    @staticmethod
    async def _open_impl_async(
        builder: ControllerBuilderPtr,
        link_builder: LinkBuilder[L],
        timeout: timedelta | None = None,  # noqa: ASYNC109
    ) -> "Controller[L]":
        runtime = Base().create_runtime()
        handle = Base().get_runtime_handle(runtime)
        timeout_ns = -1 if timeout is None else int(timeout.total_seconds() * 1000 * 1000 * 1000)
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()
        ffi_future = Base().controller_open(builder, link_builder._link_builder_ptr(), timeout_ns)
        loop.call_soon(
            lambda *_: future.set_result(Base().wait_result_controller(handle, ffi_future)),
        )
        ptr = _validate_ptr(await future)
        geometry = Base().geometry(ptr)
        link = link_builder._resolve_link(handle, ptr)
        return Controller(geometry, runtime, handle, ptr, link)

    @staticmethod
    def _open_impl(
        builder: ControllerBuilderPtr,
        link_builder: LinkBuilder[L],
        timeout: timedelta | None = None,
    ) -> "Controller[L]":
        runtime = Base().create_runtime()
        handle = Base().get_runtime_handle(runtime)
        timeout_ns = -1 if timeout is None else int(timeout.total_seconds() * 1000 * 1000 * 1000)
        ptr = _validate_ptr(
            Base().wait_result_controller(
                handle,
                Base().controller_open(builder, link_builder._link_builder_ptr(), timeout_ns),
            ),
        )
        geometry = Base().geometry(ptr)
        link = link_builder._resolve_link(handle, ptr)
        return Controller(geometry, runtime, handle, ptr, link)

    async def firmware_version_async(self: Self) -> list[FirmwareInfo]:
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()
        ffi_future = Base().controller_firmware_version_list_pointer(self._ptr)
        loop.call_soon(
            lambda *_: future.set_result(
                Base().wait_result_firmware_version_list(self._handle, ffi_future),
            ),
        )
        handle = _validate_ptr(await future)

        def get_firmware_info(i: int) -> FirmwareInfo:
            sb = ctypes.create_string_buffer(256)
            Base().controller_firmware_version_get(handle, i, sb)
            info = sb.value.decode("utf-8")
            return FirmwareInfo(info)

        res = list(map(get_firmware_info, range(self.num_devices)))
        Base().controller_firmware_version_list_pointer_delete(handle)
        return res

    def firmware_version(self: Self) -> list[FirmwareInfo]:
        handle = _validate_ptr(
            Base().wait_result_firmware_version_list(
                self._handle,
                Base().controller_firmware_version_list_pointer(self._ptr),
            ),
        )

        def get_firmware_info(i: int) -> FirmwareInfo:
            sb = ctypes.create_string_buffer(256)
            Base().controller_firmware_version_get(handle, i, sb)
            info = sb.value.decode("utf-8")
            return FirmwareInfo(info)

        res = list(map(get_firmware_info, range(self.num_devices)))
        Base().controller_firmware_version_list_pointer_delete(handle)
        return res

    async def close_async(self: Self) -> None:
        if self._disposed:
            return
        self._disposed = True
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()
        ffi_future = Base().controller_close(self._ptr)
        loop.call_soon(
            lambda *_: future.set_result(
                Base().wait_result_status(self._handle, ffi_future),
            ),
        )
        r = await future
        Base().delete_runtime(self._runtime)
        self._ptr._0 = None
        self._runtime._0 = None
        self._handle._0 = None
        _validate_status(r)

    def close(self: Self) -> None:
        if self._disposed:
            return
        self._disposed = True
        r = Base().wait_result_status(self._handle, Base().controller_close(self._ptr))
        Base().delete_runtime(self._runtime)
        self._ptr._0 = None
        self._runtime._0 = None
        self._handle._0 = None
        _validate_status(r)

    async def fpga_state_async(self: Self) -> list[FPGAState | None]:
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()
        ffi_future = Base().controller_fpga_state(self._ptr)
        loop.call_soon(
            lambda *_: future.set_result(Base().wait_result_fpga_state_list(self._handle, ffi_future)),
        )
        handle = _validate_ptr(await future)

        def get_fpga_state(i: int) -> FPGAState | None:
            state = int(Base().controller_fpga_state_get(handle, i))
            return None if state == -1 else FPGAState(state)

        res = list(map(get_fpga_state, range(self.num_devices)))
        Base().controller_fpga_state_delete(handle)
        return res

    def fpga_state(self: Self) -> list[FPGAState | None]:
        handle = _validate_ptr(
            Base().wait_result_fpga_state_list(
                self._handle,
                Base().controller_fpga_state(self._ptr),
            ),
        )

        def get_fpga_state(i: int) -> FPGAState | None:
            state = int(Base().controller_fpga_state_get(handle, i))
            return None if state == -1 else FPGAState(state)

        res = list(map(get_fpga_state, range(self.num_devices)))
        Base().controller_fpga_state_delete(handle)
        return res

    async def send_async(
        self: Self,
        d: Datagram | tuple[Datagram, Datagram],
    ) -> None:
        future: asyncio.Future = asyncio.Future()
        loop = asyncio.get_event_loop()
        ffi_future: FfiFuture
        match d:
            case Datagram():
                ffi_future = Base().controller_send(self._ptr, d._datagram_ptr(self.geometry))
            case (Datagram(), Datagram()):
                (d1, d2) = d
                d_tuple = Base().datagram_tuple(d1._datagram_ptr(self.geometry), d2._datagram_ptr(self.geometry))
                ffi_future = Base().controller_send(self._ptr, d_tuple)
            case _:
                raise InvalidDatagramTypeError
        loop.call_soon(
            lambda *_: future.set_result(Base().wait_result_status(self._handle, ffi_future)),
        )
        _validate_status(await future)

    def send(
        self: Self,
        d: Datagram | tuple[Datagram, Datagram],
    ) -> None:
        ffi_future: FfiFuture
        match d:
            case Datagram():
                ffi_future = Base().controller_send(self._ptr, d._datagram_ptr(self.geometry))
            case (Datagram(), Datagram()):
                (d1, d2) = d
                d_tuple = Base().datagram_tuple(d1._datagram_ptr(self.geometry), d2._datagram_ptr(self.geometry))
                ffi_future = Base().controller_send(self._ptr, d_tuple)
            case _:
                raise InvalidDatagramTypeError
        _validate_status(Base().wait_result_status(self._handle, ffi_future))

    def group(self: Self, group_map: Callable[[Device], K | None]) -> _GroupGuard:
        return _GroupGuard(group_map, self)

    def tick(self: "Controller[Recorder]", tick: timedelta) -> None:
        raise NotImplementedError  # pragma: no cover



from collections.abc import Iterator
from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import (
    DevicePtr,
    GeometryPtr,
)
from pyautd3.native_methods.structs import Quaternion, Vector3

from .transducer import Transducer


class Device:
    _idx: int
    _geo_ptr: GeometryPtr
    _ptr: DevicePtr
    _transducers: list[Transducer]

    def __init__(self: Self, idx: int, ptr: GeometryPtr) -> None:
        self._idx = idx
        self._geo_ptr = ptr
        self._ptr = Base().device(ptr, idx)
        self._transducers = [Transducer(i, idx, self._ptr) for i in range(int(Base().device_num_transducers(self._ptr)))]

    @property
    def idx(self: Self) -> int:
        return self._idx

    @property
    def sound_speed(self: Self) -> float:
        return float(Base().device_get_sound_speed(self._ptr))

    @sound_speed.setter
    def sound_speed(self: Self, sound_speed: float) -> None:
        Base().device_set_sound_speed(self._geo_ptr, self._idx, sound_speed)

    def set_sound_speed_from_temp(
        self: Self,
        temp: float,
        k: float = 1.4,
        r: float = 8.31446261815324,
        m: float = 28.9647e-3,
    ) -> None:
        Base().device_set_sound_speed_from_temp(self._geo_ptr, self._idx, temp, k, r, m)

    @property
    def enable(self: Self) -> bool:
        return bool(Base().device_enable_get(self._ptr))

    @enable.setter
    def enable(self: Self, value: bool) -> None:
        Base().device_enable_set(self._geo_ptr, self._idx, value)

    @property
    def num_transducers(self: Self) -> int:
        return len(self._transducers)

    @property
    def center(self: Self) -> np.ndarray:
        return Base().device_center(self._ptr).ndarray()

    def translate(self: Self, t: ArrayLike) -> None:
        Base().device_translate(self._geo_ptr, self._idx, Vector3(np.array(t)))

    def rotate(self: Self, r: ArrayLike) -> None:
        Base().device_rotate(self._geo_ptr, self._idx, Quaternion(np.array(r)))

    def affine(self: Self, t: ArrayLike, r: ArrayLike) -> None:
        Base().device_affine(self._geo_ptr, self._idx, Vector3(np.array(t)), Quaternion(np.array(r)))

    @property
    def wavelength(self: Self) -> float:
        return float(Base().device_wavelength(self._ptr))

    @property
    def wavenumber(self: Self) -> float:
        return float(Base().device_wavenumber(self._ptr))

    @property
    def rotation(self: Self) -> np.ndarray:
        return Base().device_rotation(self._ptr).ndarray()

    @property
    def x_direction(self: Self) -> np.ndarray:
        return Base().device_direction_x(self._ptr).ndarray()

    @property
    def y_direction(self: Self) -> np.ndarray:
        return Base().device_direction_y(self._ptr).ndarray()

    @property
    def axial_direction(self: Self) -> np.ndarray:
        return Base().device_direction_axial(self._ptr).ndarray()

    def __getitem__(self: Self, key: int) -> Transducer:
        return self._transducers[key]

    def __iter__(self: Self) -> Iterator[Transducer]:
        return iter(self._transducers)



from datetime import timedelta
from typing import Self

from pyautd3.driver.defined import Freq
from pyautd3.driver.defined.freq import Hz
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.native_methods.autd3_driver import SamplingConfig as _SamplingConfig
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.utils import _validate_sampling_config


class STMSamplingConfig:
    _inner: _SamplingConfig
    _n: int

    def __init__(self: Self, value: "_SamplingConfig | SamplingConfig | Freq[float] | timedelta", n: int) -> None:
        match value:
            case _SamplingConfig():
                self._inner = value
            case SamplingConfig():
                self._inner = value._inner
            case Freq():
                match value.hz:
                    case float():
                        self._inner = _validate_sampling_config(Base().stm_config_from_freq(value.hz, n))
                    case _:
                        raise TypeError
            case timedelta():
                self._inner = _validate_sampling_config(
                    Base().stm_config_from_period(int(value.total_seconds() * 1000 * 1000 * 1000), n),
                )
            case _:
                raise TypeError
        self._n = n

    @staticmethod
    def _nearest(value: "Freq[float] | timedelta", n: int) -> "STMSamplingConfig":
        match value:
            case Freq():
                match value.hz:
                    case float():
                        return STMSamplingConfig(_validate_sampling_config(Base().stm_config_from_freq_nearest(value.hz, n)), n)
                    case _:
                        raise TypeError
            case timedelta():
                return STMSamplingConfig(
                    _validate_sampling_config(
                        Base().stm_config_from_period_nearest(int(value.total_seconds() * 1000 * 1000 * 1000), n),
                    ),
                    n,
                )
            case _:
                raise TypeError

    def freq(self: Self) -> Freq[float]:
        return float(Base().stm_freq(self._inner, self._n)) * Hz

    def period(self: Self) -> timedelta:
        return timedelta(microseconds=int(Base().stm_period(self._inner, self._n)) / 1000)

    def sampling_config(self: Self) -> SamplingConfig:
        return SamplingConfig(self._inner)



import ctypes
from abc import abstractmethod
from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.native_methods.structs import Vector3

# _pad is not required for python 3.12+?


class ControlPoint(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point", Vector3),
        ("_offset", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(self: Self, point: ArrayLike) -> None:
        super().__init__()
        self._point = Vector3(np.array(point))
        self._offset = 0

    @property
    def point(self: Self) -> np.ndarray:
        return self._point.ndarray()

    @property
    def offset(self: Self) -> Phase:
        return Phase(self._offset)

    def with_phase_offset(self: Self, phase: int | Phase) -> Self:
        self._offset = Phase(phase).value
        return self


class IControlPoints:
    @staticmethod
    @abstractmethod
    def _value() -> int:
        pass


class ControlPoints1(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(self: Self, point: ArrayLike | ControlPoint) -> None:
        super().__init__()
        match point:
            case ControlPoint():
                self._point = point
            case _:
                self._point = ControlPoint(point)
        self._intensity = 0xFF

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    def with_intensity(self: Self, intensity: int | EmitIntensity) -> Self:
        self._intensity = EmitIntensity(intensity).value
        return self

    @staticmethod
    def _value() -> int:
        return 1


class ControlPoints2(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(self: Self, points: tuple[ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint]) -> None:
        super().__init__()
        match points:
            case (ControlPoint(), ControlPoint()):
                self._point1, self._point2 = points  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2 = ControlPoint(points[0]), ControlPoint(points[1])
        self._intensity = 0xFF

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    def with_intensity(self: Self, intensity: int | EmitIntensity) -> Self:
        self._intensity = EmitIntensity(intensity).value
        return self

    @staticmethod
    def _value() -> int:
        return 2


class ControlPoints3(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(self: Self, points: tuple[ArrayLike, ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint, ControlPoint]) -> None:
        super().__init__()
        match points:
            case (ControlPoint(), ControlPoint(), ControlPoint()):
                self._point1, self._point2, self._point3 = points  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3 = ControlPoint(points[0]), ControlPoint(points[1]), ControlPoint(points[2])
        self._intensity = 0xFF

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    def with_intensity(self: Self, intensity: int | EmitIntensity) -> Self:
        self._intensity = EmitIntensity(intensity).value
        return self

    @staticmethod
    def _value() -> int:
        return 3


class ControlPoints4(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        super().__init__()
        match points:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                self._point1, self._point2, self._point3, self._point4 = points  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4 = (
                    ControlPoint(points[0]),
                    ControlPoint(points[1]),
                    ControlPoint(points[2]),
                    ControlPoint(points[3]),
                )
        self._intensity = 0xFF

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    def with_intensity(self: Self, intensity: int | EmitIntensity) -> Self:
        self._intensity = EmitIntensity(intensity).value
        return self

    @staticmethod
    def _value() -> int:
        return 4


class ControlPoints5(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        super().__init__()
        match points:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                self._point1, self._point2, self._point3, self._point4, self._point5 = points  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4, self._point5 = (
                    ControlPoint(points[0]),
                    ControlPoint(points[1]),
                    ControlPoint(points[2]),
                    ControlPoint(points[3]),
                    ControlPoint(points[4]),
                )
        self._intensity = 0xFF

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    def with_intensity(self: Self, intensity: int | EmitIntensity) -> Self:
        self._intensity = EmitIntensity(intensity).value
        return self

    @staticmethod
    def _value() -> int:
        return 5


class ControlPoints6(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        super().__init__()
        match points:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                self._point1, self._point2, self._point3, self._point4, self._point5, self._point6 = points  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4, self._point5, self._point6 = (
                    ControlPoint(points[0]),
                    ControlPoint(points[1]),
                    ControlPoint(points[2]),
                    ControlPoint(points[3]),
                    ControlPoint(points[4]),
                    ControlPoint(points[5]),
                )
        self._intensity = 0xFF

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    def with_intensity(self: Self, intensity: int | EmitIntensity) -> Self:
        self._intensity = EmitIntensity(intensity).value
        return self

    @staticmethod
    def _value() -> int:
        return 6


class ControlPoints7(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_point7", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        super().__init__()
        match points:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                (
                    self._point1,
                    self._point2,
                    self._point3,
                    self._point4,
                    self._point5,
                    self._point6,
                    self._point7,
                ) = points  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4, self._point5, self._point6, self._point7 = (
                    ControlPoint(points[0]),
                    ControlPoint(points[1]),
                    ControlPoint(points[2]),
                    ControlPoint(points[3]),
                    ControlPoint(points[4]),
                    ControlPoint(points[5]),
                    ControlPoint(points[6]),
                )
        self._intensity = 0xFF

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    def with_intensity(self: Self, intensity: int | EmitIntensity) -> Self:
        self._intensity = EmitIntensity(intensity).value
        return self

    @staticmethod
    def _value() -> int:
        return 7


class ControlPoints8(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_point7", ControlPoint),
        ("_point8", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        super().__init__()
        match points:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                (
                    self._point1,
                    self._point2,
                    self._point3,
                    self._point4,
                    self._point5,
                    self._point6,
                    self._point7,
                    self._point8,
                ) = points  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4, self._point5, self._point6, self._point7, self._point8 = (
                    ControlPoint(points[0]),
                    ControlPoint(points[1]),
                    ControlPoint(points[2]),
                    ControlPoint(points[3]),
                    ControlPoint(points[4]),
                    ControlPoint(points[5]),
                    ControlPoint(points[6]),
                    ControlPoint(points[7]),
                )
        self._intensity = 0xFF

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    def with_intensity(self: Self, intensity: int | EmitIntensity) -> Self:
        self._intensity = EmitIntensity(intensity).value
        return self

    @staticmethod
    def _value() -> int:
        return 8



# This file is autogenerated
import threading
import ctypes
import os
from pyautd3.native_methods.structs import Vector3, Quaternion, FfiFuture, LocalFfiFuture
from pyautd3.native_methods.autd3_driver import SamplingConfig, LoopBehavior, SyncMode, GainSTMMode, GPIOOut, GPIOIn, Segment, SilencerTarget, Drive
from pyautd3.native_methods.autd3capi_driver import ControllerBuilderPtr, GeometryPtr, HandlePtr, LinkPtr, ResultStatus


class EmulatorControllerPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class EmulatorPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class RecordPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class SoundFieldPtr(ctypes.Structure):
    _fields_ = [("_0", ctypes.c_void_p)]


class RecordOption(ctypes.Structure):
    _fields_ = [("sound_speed", ctypes.c_float), ("time_step_ns", ctypes.c_uint64), ("print_progress", ctypes.c_bool), ("memory_limits_hint_mb", ctypes.c_uint64), ("gpu", ctypes.c_bool)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RecordOption) and self._fields_ == other._fields_  # pragma: no cover


class Range(ctypes.Structure):
    _fields_ = [("x_start", ctypes.c_float), ("x_end", ctypes.c_float), ("y_start", ctypes.c_float), ("y_end", ctypes.c_float), ("z_start", ctypes.c_float), ("z_end", ctypes.c_float), ("resolution", ctypes.c_float)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Range) and self._fields_ == other._fields_  # pragma: no cover


class ResultRecord(ctypes.Structure):
    _fields_ = [("result", RecordPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultRecord) and self._fields_ == other._fields_  # pragma: no cover


class ResultSoundField(ctypes.Structure):
    _fields_ = [("result", SoundFieldPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultSoundField) and self._fields_ == other._fields_  # pragma: no cover



class Singleton(type):
    _instances = {}  # type: ignore
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances: # pragma: no cover
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):

    def init_dll(self, bin_location: str, bin_prefix: str, bin_ext: str):
        self.dll = ctypes.CDLL(os.path.join(bin_location, f'{bin_prefix}autd3capi_emulator{bin_ext}'))

        self.dll.AUTDEmulatorTracingInit.argtypes = [] 
        self.dll.AUTDEmulatorTracingInit.restype = None

        self.dll.AUTDEmulatorTracingInitWithFile.argtypes = [ctypes.c_char_p] 
        self.dll.AUTDEmulatorTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDEmulator.argtypes = [ControllerBuilderPtr]  # type: ignore 
        self.dll.AUTDEmulator.restype = EmulatorPtr

        self.dll.AUTDEmulatorFree.argtypes = [EmulatorPtr]  # type: ignore 
        self.dll.AUTDEmulatorFree.restype = None

        self.dll.AUTDEmulatorGeometry.argtypes = [EmulatorPtr]  # type: ignore 
        self.dll.AUTDEmulatorGeometry.restype = GeometryPtr

        self.dll.AUTDEmulatorRecordFrom.argtypes = [EmulatorPtr, ctypes.c_uint64, ctypes.c_void_p]  # type: ignore 
        self.dll.AUTDEmulatorRecordFrom.restype = FfiFuture

        self.dll.AUTDEmulatorRecordFree.argtypes = [RecordPtr]  # type: ignore 
        self.dll.AUTDEmulatorRecordFree.restype = None

        self.dll.AUTDEmulatorWaitResultRecord.argtypes = [HandlePtr, FfiFuture]  # type: ignore 
        self.dll.AUTDEmulatorWaitResultRecord.restype = ResultRecord

        self.dll.AUTDEmulatorTickNs.argtypes = [LinkPtr, ctypes.c_uint64]  # type: ignore 
        self.dll.AUTDEmulatorTickNs.restype = ResultStatus

        self.dll.AUTDEmulatorRecordNumDevices.argtypes = [RecordPtr]  # type: ignore 
        self.dll.AUTDEmulatorRecordNumDevices.restype = ctypes.c_uint16

        self.dll.AUTDEmulatorRecordNumTransducers.argtypes = [RecordPtr, ctypes.c_uint16]  # type: ignore 
        self.dll.AUTDEmulatorRecordNumTransducers.restype = ctypes.c_uint8

        self.dll.AUTDEmulatorRecordDriveLen.argtypes = [RecordPtr]  # type: ignore 
        self.dll.AUTDEmulatorRecordDriveLen.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorRecordDriveTime.argtypes = [RecordPtr, ctypes.POINTER(ctypes.c_uint64)]  # type: ignore 
        self.dll.AUTDEmulatorRecordDriveTime.restype = None

        self.dll.AUTDEmulatorRecordDrivePulseWidth.argtypes = [RecordPtr, ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)]  # type: ignore 
        self.dll.AUTDEmulatorRecordDrivePulseWidth.restype = None

        self.dll.AUTDEmulatorRecordDrivePhase.argtypes = [RecordPtr, ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8)]  # type: ignore 
        self.dll.AUTDEmulatorRecordDrivePhase.restype = None

        self.dll.AUTDEmulatorRecordOutputLen.argtypes = [RecordPtr]  # type: ignore 
        self.dll.AUTDEmulatorRecordOutputLen.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorRecordOutputTime.argtypes = [RecordPtr, ctypes.POINTER(ctypes.c_uint64)]  # type: ignore 
        self.dll.AUTDEmulatorRecordOutputTime.restype = None

        self.dll.AUTDEmulatorRecordOutputVoltage.argtypes = [RecordPtr, ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_float)]  # type: ignore 
        self.dll.AUTDEmulatorRecordOutputVoltage.restype = None

        self.dll.AUTDEmulatorRecordOutputUltrasound.argtypes = [RecordPtr, ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_float)]  # type: ignore 
        self.dll.AUTDEmulatorRecordOutputUltrasound.restype = None

        self.dll.AUTDEmulatorSoundField.argtypes = [RecordPtr, Range, RecordOption]  # type: ignore 
        self.dll.AUTDEmulatorSoundField.restype = LocalFfiFuture

        self.dll.AUTDEmulatorWaitSoundField.argtypes = [HandlePtr, LocalFfiFuture]  # type: ignore 
        self.dll.AUTDEmulatorWaitSoundField.restype = ResultSoundField

        self.dll.AUTDEmulatorSoundFieldTimeLen.argtypes = [SoundFieldPtr, ctypes.c_uint64]  # type: ignore 
        self.dll.AUTDEmulatorSoundFieldTimeLen.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorSoundFieldPointsLen.argtypes = [SoundFieldPtr]  # type: ignore 
        self.dll.AUTDEmulatorSoundFieldPointsLen.restype = ctypes.c_uint64

        self.dll.AUTDEmulatorSoundFieldGetX.argtypes = [SoundFieldPtr, ctypes.POINTER(ctypes.c_float)]  # type: ignore 
        self.dll.AUTDEmulatorSoundFieldGetX.restype = None

        self.dll.AUTDEmulatorSoundFieldGetY.argtypes = [SoundFieldPtr, ctypes.POINTER(ctypes.c_float)]  # type: ignore 
        self.dll.AUTDEmulatorSoundFieldGetY.restype = None

        self.dll.AUTDEmulatorSoundFieldGetZ.argtypes = [SoundFieldPtr, ctypes.POINTER(ctypes.c_float)]  # type: ignore 
        self.dll.AUTDEmulatorSoundFieldGetZ.restype = None

        self.dll.AUTDEmulatorSoundFieldSkip.argtypes = [SoundFieldPtr, ctypes.c_uint64]  # type: ignore 
        self.dll.AUTDEmulatorSoundFieldSkip.restype = LocalFfiFuture

        self.dll.AUTDEmulatorSoundFieldNext.argtypes = [SoundFieldPtr, ctypes.c_uint64, ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(ctypes.POINTER(ctypes.c_float))]  # type: ignore 
        self.dll.AUTDEmulatorSoundFieldNext.restype = LocalFfiFuture

        self.dll.AUTDEmulatorSoundFieldFree.argtypes = [SoundFieldPtr]  # type: ignore 
        self.dll.AUTDEmulatorSoundFieldFree.restype = None

    def emulator_tracing_init(self) -> None:
        return self.dll.AUTDEmulatorTracingInit()

    def emulator_tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDEmulatorTracingInitWithFile(path)

    def emulator(self, builder: ControllerBuilderPtr) -> EmulatorPtr:
        return self.dll.AUTDEmulator(builder)

    def emulator_free(self, emulator: EmulatorPtr) -> None:
        return self.dll.AUTDEmulatorFree(emulator)

    def emulator_geometry(self, emulator: EmulatorPtr) -> GeometryPtr:
        return self.dll.AUTDEmulatorGeometry(emulator)

    def emulator_record_from(self, emulator: EmulatorPtr, start_time: int, f: ctypes.c_void_p | None) -> FfiFuture:
        return self.dll.AUTDEmulatorRecordFrom(emulator, start_time, f)

    def emulator_record_free(self, record: RecordPtr) -> None:
        return self.dll.AUTDEmulatorRecordFree(record)

    def emulator_wait_result_record(self, handle: HandlePtr, future: FfiFuture) -> ResultRecord:
        return self.dll.AUTDEmulatorWaitResultRecord(handle, future)

    def emulator_tick_ns(self, record: LinkPtr, tick_ns: int) -> ResultStatus:
        return self.dll.AUTDEmulatorTickNs(record, tick_ns)

    def emulator_record_num_devices(self, record: RecordPtr) -> ctypes.c_uint16:
        return self.dll.AUTDEmulatorRecordNumDevices(record)

    def emulator_record_num_transducers(self, record: RecordPtr, dev_idx: int) -> ctypes.c_uint8:
        return self.dll.AUTDEmulatorRecordNumTransducers(record, dev_idx)

    def emulator_record_drive_len(self, record: RecordPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorRecordDriveLen(record)

    def emulator_record_drive_time(self, record: RecordPtr, time: ctypes.Array[ctypes.c_uint64] | None) -> None:
        return self.dll.AUTDEmulatorRecordDriveTime(record, time)

    def emulator_record_drive_pulse_width(self, record: RecordPtr, dev_idx: int, tr_idx: int, pulsewidth: ctypes.Array[ctypes.c_uint8] | None) -> None:
        return self.dll.AUTDEmulatorRecordDrivePulseWidth(record, dev_idx, tr_idx, pulsewidth)

    def emulator_record_drive_phase(self, record: RecordPtr, dev_idx: int, tr_idx: int, pulsewidth: ctypes.Array[ctypes.c_uint8] | None) -> None:
        return self.dll.AUTDEmulatorRecordDrivePhase(record, dev_idx, tr_idx, pulsewidth)

    def emulator_record_output_len(self, record: RecordPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorRecordOutputLen(record)

    def emulator_record_output_time(self, record: RecordPtr, time: ctypes.Array[ctypes.c_uint64] | None) -> None:
        return self.dll.AUTDEmulatorRecordOutputTime(record, time)

    def emulator_record_output_voltage(self, record: RecordPtr, dev_idx: int, tr_idx: int, v: ctypes.Array[ctypes.c_float] | None) -> None:
        return self.dll.AUTDEmulatorRecordOutputVoltage(record, dev_idx, tr_idx, v)

    def emulator_record_output_ultrasound(self, record: RecordPtr, dev_idx: int, tr_idx: int, v: ctypes.Array[ctypes.c_float] | None) -> None:
        return self.dll.AUTDEmulatorRecordOutputUltrasound(record, dev_idx, tr_idx, v)

    def emulator_sound_field(self, record: RecordPtr, range: Range, option: RecordOption) -> LocalFfiFuture:
        return self.dll.AUTDEmulatorSoundField(record, range, option)

    def emulator_wait_sound_field(self, handle: HandlePtr, future: LocalFfiFuture) -> ResultSoundField:
        return self.dll.AUTDEmulatorWaitSoundField(handle, future)

    def emulator_sound_field_time_len(self, sound_field: SoundFieldPtr, duration_ns: int) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorSoundFieldTimeLen(sound_field, duration_ns)

    def emulator_sound_field_points_len(self, sound_field: SoundFieldPtr) -> ctypes.c_uint64:
        return self.dll.AUTDEmulatorSoundFieldPointsLen(sound_field)

    def emulator_sound_field_get_x(self, sound_field: SoundFieldPtr, x: ctypes.Array[ctypes.c_float] | None) -> None:
        return self.dll.AUTDEmulatorSoundFieldGetX(sound_field, x)

    def emulator_sound_field_get_y(self, sound_field: SoundFieldPtr, y: ctypes.Array[ctypes.c_float] | None) -> None:
        return self.dll.AUTDEmulatorSoundFieldGetY(sound_field, y)

    def emulator_sound_field_get_z(self, sound_field: SoundFieldPtr, z: ctypes.Array[ctypes.c_float] | None) -> None:
        return self.dll.AUTDEmulatorSoundFieldGetZ(sound_field, z)

    def emulator_sound_field_skip(self, sound_field: SoundFieldPtr, duration_ns: int) -> LocalFfiFuture:
        return self.dll.AUTDEmulatorSoundFieldSkip(sound_field, duration_ns)

    def emulator_sound_field_next(self, sound_field: SoundFieldPtr, duration_ns: int, time: ctypes.Array[ctypes.c_uint64] | None, v: ctypes.Array[ctypes.Array[ctypes.c_float]]) -> LocalFfiFuture:
        return self.dll.AUTDEmulatorSoundFieldNext(sound_field, duration_ns, time, v)

    def emulator_sound_field_free(self, sound_field: SoundFieldPtr) -> None:
        return self.dll.AUTDEmulatorSoundFieldFree(sound_field)



import ctypes

from pyautd3.autd_error import AUTDError
from pyautd3.native_methods.autd3_driver import SamplingConfig
from .autd3capi_driver import AUTDStatus, ResultSamplingConfig, ResultStatus
from .autd3capi import NativeMethods as Base


def _to_null_terminated_utf8(s: str) -> bytes:
    return s.encode("utf-8") + b"\0"


def _validate_status(res: ResultStatus) -> int:
    if int(res.result) == AUTDStatus.AUTDErr:
        err = ctypes.create_string_buffer(int(res.err_len))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return int(res.result)


def _validate_sampling_config(res: ResultSamplingConfig) -> SamplingConfig:
    if int(res.err_len) != 0:
        err = ctypes.create_string_buffer(int(res.err_len))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return res.result


def _validate_ptr(res):  # noqa: ANN001, ANN202
    if res.result._0 is None:
        err = ctypes.create_string_buffer(int(res.err_len))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return res.result


class ConstantADT(type):
    _initialized = False

    def __setattr__(cls, name, value):
        if cls._initialized:  # pragma: no cover
            if name in cls.__dict__:  # pragma: no cover
                raise ValueError(f"Do not assign value to {name}")  # pragma: no cover
            else:  # pragma: no cover
                raise AttributeError("Do not add new member to {cls}")  # pragma: no cover
        super().__setattr__(name, value)

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls._initialized = True



from typing import Self

from pyautd3.derive import datagram, gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga.drive import Drive
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr


@gain
@datagram
@datagram_with_segment
class Uniform(Gain):
    _drive: Drive

    def __init__(self: Self, drive: Drive | EmitIntensity | Phase | tuple) -> None:
        super().__init__()
        self._drive = Drive(drive)

    @property
    def drive(self: Self) -> Drive:
        return self._drive

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        return Base().gain_uniform(self._drive._intensity.value, self._drive._phase.value)



import ctypes
from collections.abc import Callable
from typing import Self

from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Device, Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GeometryPtr


@datagram
class ReadsFPGAState(Datagram):
    def __init__(self: Self, f: Callable[[Device], bool]) -> None:
        super().__init__()

        def f_native(_context: ctypes.c_void_p, geometry_ptr: GeometryPtr, dev_idx: int) -> bool:
            return f(Device(dev_idx, geometry_ptr))

        self._f_native = ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_void_p, GeometryPtr, ctypes.c_uint16)(f_native)

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().datagram_reads_fpga_state(self._f_native, None, geometry._geometry_ptr)  # type: ignore[arg-type]



from typing import Self

from pyautd3.derive import builder
from pyautd3.driver.link import Link, LinkBuilder
from pyautd3.native_methods.autd3capi import ControllerPtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import HandlePtr, LinkBuilderPtr, LinkPtr
from pyautd3.native_methods.autd3capi_link_twincat import NativeMethods as LinkTwinCAT
from pyautd3.native_methods.utils import _to_null_terminated_utf8, _validate_ptr


class _TwinCATBuilder(LinkBuilder["TwinCAT"]):
    def _resolve_link(self: Self, handle: HandlePtr, ptr: ControllerPtr) -> "TwinCAT":
        return TwinCAT(handle, Base().link_get(ptr))  # pragma: no cover

    def _link_builder_ptr(self: LinkBuilder) -> LinkBuilderPtr:
        return LinkTwinCAT().link_twin_cat()  # pragma: no cover


class TwinCAT(Link):
    @staticmethod
    def builder() -> _TwinCATBuilder:
        return _TwinCATBuilder()

    def __init__(self: Self, handle: HandlePtr, ptr: LinkPtr) -> None:
        super().__init__(handle, ptr)  # pragma: no cover


@builder
class _RemoteTwinCATBuilder(LinkBuilder["RemoteTwinCAT"]):
    _prop_server_ams_net_id: str
    _param_server_ip: str
    _param_client_ams_net_id: str

    def __init__(self: Self, server_ams_net_id: str) -> None:
        self._prop_server_ams_net_id = server_ams_net_id
        self._param_server_ip = ""
        self._param_client_ams_net_id = ""

    def _link_builder_ptr(self: Self) -> LinkBuilderPtr:
        return _validate_ptr(  # pragma: no cover
            LinkTwinCAT().link_remote_twin_cat(
                _to_null_terminated_utf8(self._prop_server_ams_net_id),
                _to_null_terminated_utf8(self._param_server_ip),
                _to_null_terminated_utf8(self._param_client_ams_net_id),
            ),
        )

    def _resolve_link(self: Self, handle: HandlePtr, _ptr: ControllerPtr) -> "RemoteTwinCAT":
        return RemoteTwinCAT(handle, Base().link_get(_ptr))  # pragma: no cover


class RemoteTwinCAT(Link):
    def __init__(self: Self, handle: HandlePtr, ptr: LinkPtr) -> None:
        super().__init__(handle, ptr)  # pragma: no cover

    @staticmethod
    def builder(server_ams_net_id: str) -> _RemoteTwinCATBuilder:
        return _RemoteTwinCATBuilder(server_ams_net_id)



from typing import Self

from pyautd3.derive import datagram, gain
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr


@gain
@datagram
@datagram_with_segment
class Null(Gain):
    def __init__(self: Self) -> None:
        super().__init__()

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        return Base().gain_null()



import ctypes
from collections.abc import Iterable
from datetime import timedelta
from typing import Self

import numpy as np

from pyautd3.derive import datagram, modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.modulation.resample import Resampler
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr


@datagram
@datagram_with_segment
@modulation
class Custom(Modulation):
    _buf: np.ndarray
    _config: SamplingConfig | tuple[Freq[float], SamplingConfig, Resampler]

    def __init__(self: Self, buf: Iterable[int], config: SamplingConfig | Freq[int] | Freq[float] | timedelta) -> None:
        super().__init__()
        self._buf = np.fromiter(buf, dtype=np.uint8)
        self._config = config

    @staticmethod
    def new_with_resample(
        buf: Iterable[int],
        source: Freq[float],
        target: SamplingConfig | Freq[int] | Freq[float] | timedelta,
        resampler: Resampler,
    ) -> "Custom":
        instance = Custom(buf, SamplingConfig(target))
        instance._config = (source, SamplingConfig(target), resampler)
        return instance

    def _modulation_ptr(self: Self) -> ModulationPtr:
        match self._config:
            case (Freq(), SamplingConfig(), Resampler()):
                (source, target, resampler) = self._config
                return Base().modulation_custom_with_resample(
                    self._loop_behavior,
                    self._buf.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                    len(self._buf),
                    source.hz,
                    target._inner,
                    resampler._dyn_resampler(),
                )
            case _:
                return Base().modulation_custom(
                    self._config._inner,
                    self._loop_behavior,
                    self._buf.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
                    len(self._buf),
                )



from typing import Self
from pyautd3.derive import builder
from pyautd3.driver.link import Link
from pyautd3.driver.link import LinkBuilder
from pyautd3.native_methods.autd3capi import ControllerPtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import HandlePtr
from pyautd3.native_methods.autd3capi_driver import LinkBuilderPtr
from pyautd3.native_methods.autd3capi_driver import LinkPtr
from pyautd3.native_methods.autd3capi_link_twincat import NativeMethods as LinkTwinCAT
from pyautd3.native_methods.utils import _to_null_terminated_utf8
from pyautd3.native_methods.utils import _validate_ptr



class _TwinCATBuilder(LinkBuilder[TwinCAT]):
    def _resolve_link(self: _TwinCATBuilder, handle: HandlePtr, ptr: ControllerPtr) -> TwinCAT: ...
    def _link_builder_ptr(self: _TwinCATBuilder, ) -> LinkBuilderPtr: ...

class TwinCAT(Link):
    def __init__(self: TwinCAT, handle: HandlePtr, ptr: LinkPtr) -> None: ...
    @staticmethod
    def builder() -> _TwinCATBuilder: ...

class _RemoteTwinCATBuilder(LinkBuilder[RemoteTwinCAT]):
    def __init__(self: _RemoteTwinCATBuilder, server_ams_net_id: str) -> None: ...
    def _link_builder_ptr(self: _RemoteTwinCATBuilder, ) -> LinkBuilderPtr: ...
    def _resolve_link(self: _RemoteTwinCATBuilder, handle: HandlePtr, _ptr: ControllerPtr) -> RemoteTwinCAT: ...
    def with_server_ip(self: _RemoteTwinCATBuilder, server_ip: str) -> _RemoteTwinCATBuilder: ...
    def with_client_ams_net_id(self: _RemoteTwinCATBuilder, client_ams_net_id: str) -> _RemoteTwinCATBuilder: ...
    @property
    def server_ams_net_id(self: _RemoteTwinCATBuilder) -> str: ...
    @property
    def server_ip(self: _RemoteTwinCATBuilder) -> str: ...
    @property
    def client_ams_net_id(self: _RemoteTwinCATBuilder) -> str: ...

class RemoteTwinCAT(Link):
    def __init__(self: RemoteTwinCAT, handle: HandlePtr, ptr: LinkPtr) -> None: ...
    @staticmethod
    def builder(server_ams_net_id: str) -> _RemoteTwinCATBuilder: ...



from typing import Self
from pyautd3.derive import datagram
from pyautd3.derive import modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.utils import _validate_u8
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.modulation.cache import Cache
from pyautd3.modulation.fir import Fir
from collections.abc import Iterable
from pyautd3.modulation.radiation_pressure import RadiationPressure
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class Static(Modulation):
    _intensity: int
    def __init__(self: Static, intensity: int | None = None) -> None: ...
    def _modulation_ptr(self: Static, ) -> ModulationPtr: ...
    def with_cache(self: Static, ) -> Cache[Static]: ...
    def with_fir(self: Static, iterable: Iterable[float]) -> Fir[Static]: ...
    def with_radiation_pressure(self: Static, ) -> RadiationPressure[Static]: ...
    def with_timeout(self: Static, timeout: timedelta | None) -> DatagramWithTimeout[Static]: ...
    def with_parallel_threshold(self: Static, threshold: int | None) -> DatagramWithParallelThreshold[Static]: ...
    def with_segment(self: Static, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Static]: ...
    @staticmethod
    def with_intensity(intensity: int) -> Static: ...
    @property
    def intensity(self: Static) -> int: ...



from typing import Self

import numpy as np

from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import (
    DevicePtr,
    TransducerPtr,
)


class Transducer:
    _idx: int
    _dev_idx: int
    _ptr: TransducerPtr

    def __init__(self: Self, idx: int, dev_idx: int, ptr: DevicePtr) -> None:
        self._idx = idx
        self._dev_idx = dev_idx
        self._ptr = Base().transducer(ptr, idx)

    @property
    def idx(self: Self) -> int:
        return self._idx

    @property
    def dev_idx(self: Self) -> int:
        return self._dev_idx

    @property
    def position(self: Self) -> np.ndarray:
        return Base().transducer_position(self._ptr).ndarray()



import ctypes
from collections.abc import Iterable
from datetime import timedelta
from typing import Self

import numpy as np

from pyautd3.derive import datagram
from pyautd3.derive.derive_builder import builder
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.datagram.stm.stm_sampling_config import STMSamplingConfig
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga import LoopBehavior
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GainPtr, GainSTMMode, GainSTMPtr, Segment, TransitionModeWrap
from pyautd3.native_methods.autd3capi_driver import LoopBehavior as _LoopBehavior
from pyautd3.native_methods.utils import _validate_ptr

__all__ = []  # type: ignore[var-annotated]


@builder
@datagram
@datagram_with_segment
class GainSTM(DatagramS[GainSTMPtr], Datagram):
    _gains: np.ndarray
    _param_mode: GainSTMMode

    _stm_sampling_config: STMSamplingConfig
    _param_loop_behavior: _LoopBehavior

    def __private_init__(
        self: Self,
        sampling_config: STMSamplingConfig,
        gains: list[Gain],
    ) -> None:
        self._gains = np.array(gains)
        self._param_mode = GainSTMMode.PhaseIntensityFull

        self._stm_sampling_config = sampling_config
        self._param_loop_behavior = LoopBehavior.Infinite

    def __init__(
        self: Self,
        config: "SamplingConfig | Freq[float] | timedelta",
        iterable: Iterable[Gain],
    ) -> None:
        gains = list(iterable)
        self.__private_init__(STMSamplingConfig(config, len(gains)), gains)

    @classmethod
    def nearest(
        cls: type["GainSTM"],
        config: "Freq[float] | timedelta",
        iterable: Iterable[Gain],
    ) -> "GainSTM":
        ins = cls.__new__(cls)
        gains = list(iterable)
        ins.__private_init__(STMSamplingConfig._nearest(config, len(gains)), gains)
        return ins

    def _raw_ptr(self: Self, geometry: Geometry) -> GainSTMPtr:
        gains: np.ndarray = np.ndarray(len(self._gains), dtype=GainPtr)
        for i, g in enumerate(self._gains):
            gains[i]["_0"] = g._gain_ptr(geometry)._0
        return self._ptr(gains)

    def _ptr(self: Self, gains: np.ndarray) -> GainSTMPtr:
        return _validate_ptr(
            Base().stm_gain(
                self._stm_sampling_config._inner,
                gains.ctypes.data_as(ctypes.POINTER(GainPtr)),  # type: ignore[arg-type]
                len(gains),
                self._param_mode,
                self._param_loop_behavior,
            ),
        )

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().stm_gain_into_datagram(self._raw_ptr(geometry))

    def _into_segment(self: Self, ptr: GainSTMPtr, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr:
        return Base().stm_gain_into_datagram_with_segment(
            ptr,
            segment,
            transition_mode if transition_mode is not None else TransitionMode.NONE,
        )

    @property
    def freq(self: Self) -> Freq[float]:
        return self._stm_sampling_config.freq()

    @property
    def period(self: Self) -> timedelta:
        return self._stm_sampling_config.period()

    @property
    def sampling_config(self: Self) -> SamplingConfig:
        return self._stm_sampling_config.sampling_config()

    def _sampling_config_intensity(self: Self) -> SamplingConfig:
        return self.sampling_config

    def _sampling_config_phase(self: Self) -> SamplingConfig:
        return self.sampling_config



from abc import ABCMeta, abstractmethod
from typing import Self

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GainPtr, Segment, TransitionModeWrap

__all__ = []  # type: ignore[var-annotated]


class Gain(
    DatagramS[GainPtr],
    Datagram,
    metaclass=ABCMeta,
):
    def __init__(self: Self) -> None:
        super().__init__()

    def _raw_ptr(self: Self, geometry: Geometry) -> GainPtr:
        return self._gain_ptr(geometry)

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().gain_into_datagram(self._gain_ptr(geometry))

    def _into_segment(self: Self, ptr: GainPtr, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr:
        return Base().gain_into_datagram_with_segment(
            ptr,
            segment,
            transition_mode if transition_mode is not None else TransitionMode.NONE,
        )

    @abstractmethod
    def _gain_ptr(self: Self, geometry: Geometry) -> GainPtr:
        pass



from abc import ABCMeta, abstractmethod
from typing import Generic, Self, TypeVar

from pyautd3.native_methods.autd3capi import ControllerPtr
from pyautd3.native_methods.autd3capi_driver import HandlePtr, LinkBuilderPtr, LinkPtr

__all__ = []  # type: ignore[var-annotated]

L = TypeVar("L", bound="Link")


class Link(metaclass=ABCMeta):
    _ptr: LinkPtr
    _handle: HandlePtr

    def __init__(self: Self, handle: HandlePtr, ptr: LinkPtr) -> None:
        self._ptr = ptr
        self._handle = handle


class LinkBuilder(Generic[L], metaclass=ABCMeta):
    @abstractmethod
    def _link_builder_ptr(self: Self) -> LinkBuilderPtr:
        pass

    @abstractmethod
    def _resolve_link(self: Self, _handle: HandlePtr, _ptr: ControllerPtr) -> L:
        pass



from collections.abc import Iterable
from typing import Self

from pyautd3.derive import builder, datagram, modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.modulation.sine import Sine
from pyautd3.native_methods.autd3capi_driver import ModulationPtr


@modulation
@datagram
@datagram_with_segment
@builder
class Fourier(Modulation):
    _components: list[Sine]
    _param_clamp: bool
    _param_scale_factor: float | None
    _param_offset: int

    def __init__(self: Self, iterable: Iterable[Sine]) -> None:
        super().__init__()
        self._components = list(iterable)
        self._param_clamp = False
        self._param_scale_factor = float("nan")
        self._param_offset = 0

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return self._components[0]._mode.fourier_ptr(
            self._components,
            len(self._components),
            self._param_clamp,
            self._param_scale_factor if self._param_scale_factor is not None else float("nan"),  # type: ignore[arg-type]
            self._param_offset,
            self._loop_behavior,
        )



from abc import ABCMeta
from abc import abstractmethod
from typing import Generic
from typing import Self
from typing import TypeVar
from pyautd3.derive import datagram
from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi_driver import DatagramPtr
from pyautd3.native_methods.autd3capi_driver import Segment
from pyautd3.native_methods.autd3capi_driver import TransitionModeWrap
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold

DS = TypeVar("DS", bound="DatagramS")
P = TypeVar("P")

class DatagramS(Generic[P]):
    def _into_segment(self: DatagramS[P], ptr: P, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramPtr: ...
    def _raw_ptr(self: DatagramS[P], geometry: Geometry) -> P: ...

class DatagramWithSegment(Datagram, Generic[DS]):
    _datagram: DS
    _segment: Segment
    _transition_mode: TransitionModeWrap | None
    def __init__(self: DatagramWithSegment[DS], datagram: DS, segment: Segment, transition_mode: TransitionModeWrap | None) -> None: ...
    def _datagram_ptr(self: DatagramWithSegment[DS], g: Geometry) -> DatagramPtr: ...
    def with_timeout(self: DatagramWithSegment[DS], timeout: timedelta | None) -> DatagramWithTimeout[DatagramWithSegment[DS]]: ...
    def with_parallel_threshold(self: DatagramWithSegment[DS], threshold: int | None) -> DatagramWithParallelThreshold[DatagramWithSegment[DS]]: ...



from datetime import timedelta
from pathlib import Path
from typing import Self
from pyautd3 import SamplingConfig
from pyautd3.derive import datagram
from pyautd3.derive import modulation
from pyautd3.derive.derive_datagram import datagram_with_segment
from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.defined.freq import Freq
from pyautd3.modulation.resample import Resampler
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.native_methods.autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile
from pyautd3.native_methods.utils import _to_null_terminated_utf8
from pyautd3.native_methods.utils import _validate_ptr
from pyautd3.modulation.cache import Cache
from pyautd3.modulation.fir import Fir
from collections.abc import Iterable
from pyautd3.modulation.radiation_pressure import RadiationPressure
from datetime import timedelta
from pyautd3.driver.datagram.with_timeout import DatagramWithTimeout
from pyautd3.driver.datagram.with_parallel_threshold import DatagramWithParallelThreshold
from pyautd3.native_methods.autd3capi_driver import Segment, TransitionModeWrap
from pyautd3.driver.datagram.with_segment import DatagramWithSegment



class Wav(Modulation):
    _path: Path
    _resampler: tuple[SamplingConfig, Resampler] | None
    def __init__(self: Wav, path: Path) -> None: ...
    def _modulation_ptr(self: Wav, ) -> ModulationPtr: ...
    def with_cache(self: Wav, ) -> Cache[Wav]: ...
    def with_fir(self: Wav, iterable: Iterable[float]) -> Fir[Wav]: ...
    def with_radiation_pressure(self: Wav, ) -> RadiationPressure[Wav]: ...
    def with_timeout(self: Wav, timeout: timedelta | None) -> DatagramWithTimeout[Wav]: ...
    def with_parallel_threshold(self: Wav, threshold: int | None) -> DatagramWithParallelThreshold[Wav]: ...
    def with_segment(self: Wav, segment: Segment, transition_mode: TransitionModeWrap | None) -> DatagramWithSegment[Wav]: ...
    @staticmethod
    def new_with_resample(path: Path, target: SamplingConfig | Freq[int] | Freq[float] | timedelta, resampler: Resampler) -> Wav: ...
